from __future__ import annotations

import datetime

import numpy as np
import pandas as pd
import pyarrow as pa
import pytest

from daft import col
from daft.recordbatch import MicroPartition
from daft.series import Series


class CustomClass:
    def __init__(self, x):
        self.x = x

    def __eq__(self, other):
        return self.x == other.x

    def __hash__(self):
        return hash(self.x)


class CustomClassWithoutHash:
    def __init__(self, x):
        self.x = x

    def __eq__(self, other):
        return self.x == other.x


@pytest.mark.parametrize(
    "input,items,expected",
    [
        pytest.param([None, None], [None], [None, None], id="NullColumn"),
        pytest.param([True, False], [False], [False, True], id="BooleanColumn"),
        pytest.param(["a", "b", "c", "d"], ["a", "b"], [True, True, False, False], id="StringColumn"),
        pytest.param([b"a", b"b", b"c", b"d"], [b"a", b"b"], [True, True, False, False], id="BinaryColumn"),
        pytest.param([-1, 2, 3, 4], [-1, 2], [True, True, False, False], id="IntColumn"),
        pytest.param([-1.0, 2.0, 3.0, 4.0], [-1.0, 2.0], [True, True, False, False], id="FloatColumn"),
        pytest.param(
            [datetime.date.today(), datetime.date.today() - datetime.timedelta(days=1)],
            [datetime.date.today()],
            [True, False],
            id="DateColumn",
        ),
        pytest.param(
            [datetime.datetime(2022, 1, 1), datetime.datetime(2023, 1, 1)],
            [datetime.datetime(2022, 1, 1)],
            [True, False],
            id="TimestampColumn",
        ),
        pytest.param([CustomClass(1), CustomClass(2)], [CustomClass(1)], [True, False], id="ObjectColumn"),
        pytest.param(
            [CustomClassWithoutHash(1), CustomClassWithoutHash(2)],
            [CustomClassWithoutHash(1)],
            [True, False],
            id="ObjectWithoutHashColumn",
        ),
    ],
)
def test_table_expr_is_in_same_types(input, items, expected) -> None:
    daft_recordbatch = MicroPartition.from_pydict({"input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").is_in(items)])
    pydict = daft_recordbatch.to_pydict()

    assert pydict["input"] == expected


@pytest.mark.parametrize(
    "input,items,expected",
    [
        # Int
        pytest.param([-1, 2, 3, 4], ["-1", "2"], None, id="IntWithString"),
        pytest.param([1, 2, 3, 4], [1.0, 2.0], [True, True, False, False], id="IntWithFloat"),
        pytest.param([0, 1, 2, 3], [True], [False, True, False, False], id="IntWithBool"),
        # Float
        pytest.param([-1.0, 2.0, 3.0, 4.0], ["-1.0", "2.0"], None, id="FloatWithString"),
        pytest.param([1.0, 2.0, 3.0, 4.0], [1, 2], [True, True, False, False], id="FloatWithInt"),
        pytest.param([0.0, 1.0, 2.0, 3.0], [True], [False, True, False, False], id="FloatWithBool"),
        # String
        pytest.param(["1", "2", "3", "4"], [1, 2], None, id="StringWithInt"),
        pytest.param(["1.0", "2.0", "3.0", "4.0"], [1.0, 2.0], None, id="StringWithFloat"),
        # Bool
        pytest.param([True, False, None], [1, 0], [True, True, None], id="BoolWithInt"),
        pytest.param([True, False, None], [1.0], [True, False, None], id="BoolWithFloat"),
        # Date
        pytest.param(
            [datetime.date.today(), datetime.date.today() - datetime.timedelta(days=1)],
            [datetime.datetime.today()],
            [True, False],
            id="DateWithTimestamp",
        ),
        # Timestamp
        pytest.param(
            [datetime.datetime(2022, 1, 1), datetime.datetime(2023, 1, 1)],
            [datetime.date(2022, 1, 1)],
            [True, False],
            id="TimestampWithDate",
        ),
    ],
)
def test_table_expr_is_in_different_types_castable(input, items, expected) -> None:
    daft_recordbatch = MicroPartition.from_pydict({"input": input})

    if expected is None:
        with pytest.raises(ValueError, match="Cannot perform comparison on types:"):
            daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").is_in(items)])
    else:
        daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").is_in(items)])
        pydict = daft_recordbatch.to_pydict()
        assert pydict["input"] == expected


@pytest.mark.parametrize(
    "input,items,expected",
    [
        pytest.param([None, None, None], [None], [None, None, None], id="NullColumn"),
        pytest.param([True, False, None], [None], [False, False, None], id="BooleanColumn"),
        pytest.param(["a", "b", None], [None], [False, False, None], id="StringColumn"),
        pytest.param([b"a", b"b", None], [None], [False, False, None], id="BinaryColumn"),
        pytest.param([1, 2, None], [None], [False, False, None], id="IntColumn"),
        pytest.param([1.0, 2.0, None], [None], [False, False, None], id="FloatColumn"),
    ],
)
def test_table_expr_is_in_items_is_none(input, items, expected) -> None:
    daft_recordbatch = MicroPartition.from_pydict({"input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").is_in(items)])
    pydict = daft_recordbatch.to_pydict()

    assert pydict["input"] == expected


@pytest.mark.parametrize(
    "input,items,expected",
    [
        pytest.param([1, 2, 3, 4], np.array([1, 2]), [True, True, False, False], id="NumpyArray"),
        pytest.param([1, 2, 3, 4], pa.array([1, 2], type=pa.int8()), [True, True, False, False], id="PyArrowArray"),
        pytest.param([1, 2, 3, 4], pd.Series([1, 2]), [True, True, False, False], id="PandasSeries"),
        pytest.param([1, 2, 3, 4], Series.from_pylist([1, 2]), [True, True, False, False], id="DaftSeries"),
    ],
)
def test_table_expr_is_in_different_input_types(input, items, expected) -> None:
    daft_recordbatch = MicroPartition.from_pydict({"input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").is_in(items)])
    pydict = daft_recordbatch.to_pydict()

    assert pydict["input"] == expected


def test_table_expr_is_in_with_another_df_column() -> None:
    daft_recordbatch = MicroPartition.from_pydict({"input": [1, 2, 3, 4], "items": [3, 4, 5, 6]})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").is_in(col("items"))])
    pydict = daft_recordbatch.to_pydict()

    assert pydict["input"] == [False, False, True, True]


def test_table_expr_is_in_empty_items() -> None:
    daft_recordbatch = MicroPartition.from_pydict({"input": [1, 2, 3, 4]})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").is_in([])])
    pydict = daft_recordbatch.to_pydict()

    assert pydict["input"] == [False, False, False, False]


def test_table_expr_is_in_items_invalid_input() -> None:
    daft_recordbatch = MicroPartition.from_pydict({"input": [1, 2, 3, 4]})

    with pytest.raises(ValueError, match="Creating a Series from data of type"):
        daft_recordbatch.eval_expression_list([col("input").is_in(1)])



from __future__ import annotations

import contextlib
import dataclasses
import logging
from typing import TYPE_CHECKING, Any, ClassVar

from daft.daft import IOConfig, PyDaftContext, PyDaftExecutionConfig, PyDaftPlanningConfig
from daft.daft import get_context as _get_context
from daft.daft import set_runner_native as _set_runner_native
from daft.daft import set_runner_ray as _set_runner_ray

if TYPE_CHECKING:
    from collections.abc import Generator

    from daft.runners.partitioning import PartitionT
    from daft.runners.runner import Runner

logger = logging.getLogger(__name__)

import threading


@dataclasses.dataclass
class DaftContext:
    """Global context for the current Daft execution environment."""

    _ctx: PyDaftContext

    _lock: ClassVar[threading.Lock] = threading.Lock()

    @property
    def _runner(self) -> Runner[PartitionT]:
        return self._ctx._runner

    @_runner.setter
    def _runner(self, runner: Runner[PartitionT]) -> None:
        self._ctx._runner = runner

    @staticmethod
    def _from_native(ctx: PyDaftContext) -> DaftContext:
        return DaftContext(ctx=ctx)

    def __init__(self, ctx: PyDaftContext | None = None):
        if ctx is not None:
            self._ctx = ctx
        else:
            self._ctx = PyDaftContext()

    def get_or_infer_runner_type(self) -> str:
        """Get or infer the runner type.

        This API will get or infer the currently used runner type according to the following strategies:
        1. If the `runner` has been set, return its type directly;
        2. Try to determine whether it's currently running on a ray cluster. If so, consider it to be a ray type;
        3. Try to determine based on `DAFT_RUNNER` env variable.

        :return: runner type string ("native" or "ray")
        """
        if self._ctx._runner is not None:
            return self._ctx._runner.name

        return self._ctx.get_or_infer_runner_type()

    def get_or_create_runner(self) -> Runner[PartitionT]:
        return self._ctx.get_or_create_runner()

    @property
    def daft_execution_config(self) -> PyDaftExecutionConfig:
        return self._ctx._daft_execution_config

    @property
    def daft_planning_config(self) -> PyDaftPlanningConfig:
        return self._ctx._daft_planning_config


def get_context() -> DaftContext:
    return DaftContext(_get_context())


def set_runner_ray(
    address: str | None = None,
    noop_if_initialized: bool = False,
    max_task_backlog: int | None = None,
    force_client_mode: bool = False,
) -> DaftContext:
    """Configure Daft to execute dataframes using the Ray distributed computing framework.

    Args:
        address: Ray cluster address to connect to. If None, connects to or starts a local Ray instance.
        noop_if_initialized: If True, skip initialization if Ray is already running.
        max_task_backlog: Maximum number of tasks that can be queued. None means Daft will automatically determine a good default.
        force_client_mode: If True, forces Ray to run in client mode.

    Returns:
        DaftContext: Updated Daft execution context configured for Ray.

    Note:
        Can also be configured via environment variable: DAFT_RUNNER=ray
    """
    py_ctx = _set_runner_ray(
        address=address,
        noop_if_initialized=noop_if_initialized,
        max_task_backlog=max_task_backlog,
        force_client_mode=force_client_mode,
    )

    return DaftContext._from_native(py_ctx)


def set_runner_native(num_threads: int | None = None) -> DaftContext:
    """Configure Daft to execute dataframes using native multi-threaded processing.

    This is the default execution mode for Daft.

    Returns:
        DaftContext: Updated Daft execution context configured for native execution.

    Note:
        Can also be configured via environment variable: DAFT_RUNNER=native
    """
    py_ctx = _set_runner_native(num_threads=num_threads)

    return DaftContext._from_native(py_ctx)


@contextlib.contextmanager
def planning_config_ctx(**kwargs: Any) -> Generator[None, None, None]:
    """Context manager that wraps set_planning_config to reset the config to its original setting afternwards."""
    original_config = get_context().daft_planning_config
    try:
        set_planning_config(**kwargs)
        yield
    finally:
        set_planning_config(config=original_config)


def set_planning_config(
    config: PyDaftPlanningConfig | None = None,
    default_io_config: IOConfig | None = None,
) -> DaftContext:
    """Globally sets various configuration parameters which control Daft plan construction behavior.

    These configuration values are used when a Dataframe is being constructed (e.g. calls to create a Dataframe, or to build on an existing Dataframe).

    Args:
        config: A PyDaftPlanningConfig object to set the config to, before applying other kwargs. Defaults to None which indicates
            that the old (current) config should be used.
        default_io_config: A default IOConfig to use in the absence of one being explicitly passed into any Expression (e.g. `.url.download()`)
            or Dataframe operation (e.g. `daft.read_parquet()`).
    """
    # Replace values in the DaftPlanningConfig with user-specified overrides
    ctx = get_context()
    with ctx._lock:
        old_daft_planning_config = ctx._ctx._daft_planning_config if config is None else config
        new_daft_planning_config = old_daft_planning_config.with_config_values(
            default_io_config=default_io_config,
        )

        ctx._ctx._daft_planning_config = new_daft_planning_config
        return ctx


@contextlib.contextmanager
def execution_config_ctx(**kwargs: Any) -> Generator[None, None, None]:
    """Context manager that wraps set_execution_config to reset the config to its original setting afternwards."""
    original_config = get_context()._ctx._daft_execution_config
    try:
        set_execution_config(**kwargs)
        yield
    finally:
        set_execution_config(config=original_config)


def set_execution_config(
    config: PyDaftExecutionConfig | None = None,
    scan_tasks_min_size_bytes: int | None = None,
    scan_tasks_max_size_bytes: int | None = None,
    max_sources_per_scan_task: int | None = None,
    broadcast_join_size_bytes_threshold: int | None = None,
    parquet_split_row_groups_max_files: int | None = None,
    sort_merge_join_sort_with_aligned_boundaries: bool | None = None,
    hash_join_partition_size_leniency: float | None = None,
    sample_size_for_sort: int | None = None,
    num_preview_rows: int | None = None,
    parquet_target_filesize: int | None = None,
    parquet_target_row_group_size: int | None = None,
    parquet_inflation_factor: float | None = None,
    csv_target_filesize: int | None = None,
    csv_inflation_factor: float | None = None,
    shuffle_aggregation_default_partitions: int | None = None,
    partial_aggregation_threshold: int | None = None,
    high_cardinality_aggregation_threshold: float | None = None,
    read_sql_partition_size_bytes: int | None = None,
    enable_aqe: bool | None = None,
    default_morsel_size: int | None = None,
    shuffle_algorithm: str | None = None,
    pre_shuffle_merge_threshold: int | None = None,
    flight_shuffle_dirs: list[str] | None = None,
    enable_ray_tracing: bool | None = None,
    scantask_splitting_level: int | None = None,
    native_parquet_writer: bool | None = None,
    use_experimental_distributed_engine: bool | None = None,
    min_cpu_per_task: float | None = None,
) -> DaftContext:
    """Globally sets various configuration parameters which control various aspects of Daft execution.

    These configuration values
    are used when a Dataframe is executed (e.g. calls to `DataFrame.write_*`, [DataFrame.collect()](https://docs.getdaft.io/en/stable/api/dataframe/#daft.DataFrame.collect) or [DataFrame.show()](https://docs.getdaft.io/en/stable/api/dataframe/#daft.DataFrame.select)).

    Args:
        config: A PyDaftExecutionConfig object to set the config to, before applying other kwargs. Defaults to None which indicates
            that the old (current) config should be used.
        scan_tasks_min_size_bytes: Minimum size in bytes when merging ScanTasks when reading files from storage.
            Increasing this value will make Daft perform more merging of files into a single partition before yielding,
            which leads to bigger but fewer partitions. (Defaults to 96 MiB)
        scan_tasks_max_size_bytes: Maximum size in bytes when merging ScanTasks when reading files from storage.
            Increasing this value will increase the upper bound of the size of merged ScanTasks, which leads to bigger but
            fewer partitions. (Defaults to 384 MiB)
        max_sources_per_scan_task: Maximum number of sources in a single ScanTask. (Defaults to 10)
        broadcast_join_size_bytes_threshold: If one side of a join is smaller than this threshold, a broadcast join will be used.
            Default is 10 MiB.
        parquet_split_row_groups_max_files: Maximum number of files to read in which the row group splitting should happen. (Defaults to 10)
        sort_merge_join_sort_with_aligned_boundaries: Whether to use a specialized algorithm for sorting both sides of a
            sort-merge join such that they have aligned boundaries. This can lead to a faster merge-join at the cost of
            more skewed sorted join inputs, increasing the risk of OOMs.
        hash_join_partition_size_leniency: If the left side of a hash join is already correctly partitioned and the right side isn't,
            and the ratio between the left and right size is at least this value, then the right side is repartitioned to have an equal
            number of partitions as the left. Defaults to 0.5.
        sample_size_for_sort: number of elements to sample from each partition when running sort,
            Default is 20.
        num_preview_rows: number of rows to when showing a dataframe preview,
            Default is 8.
        parquet_target_filesize: Target File Size when writing out Parquet Files. Defaults to 512MB
        parquet_target_row_group_size: Target Row Group Size when writing out Parquet Files. Defaults to 128MB
        parquet_inflation_factor: Inflation Factor of parquet files (In-Memory-Size / File-Size) ratio. Defaults to 3.0
        csv_target_filesize: Target File Size when writing out CSV Files. Defaults to 512MB
        csv_inflation_factor: Inflation Factor of CSV files (In-Memory-Size / File-Size) ratio. Defaults to 0.5
        shuffle_aggregation_default_partitions: Maximum number of partitions to create when performing aggregations on the Ray Runner. Defaults to 200, unless the number of input partitions is less than 200.
        partial_aggregation_threshold: Threshold for performing partial aggregations on the Native Runner. Defaults to 10000 rows.
        high_cardinality_aggregation_threshold: Threshold selectivity for performing high cardinality aggregations on the Native Runner. Defaults to 0.8.
        read_sql_partition_size_bytes: Target size of partition when reading from SQL databases. Defaults to 512MB
        enable_aqe: Enables Adaptive Query Execution, Defaults to False
        default_morsel_size: Default size of morsels used for the new local executor. Defaults to 131072 rows.
        shuffle_algorithm: The shuffle algorithm to use. Defaults to "auto", which will let Daft determine the algorithm. Options are "map_reduce" and "pre_shuffle_merge".
        pre_shuffle_merge_threshold: Memory threshold in bytes for pre-shuffle merge. Defaults to 1GB
        flight_shuffle_dirs: The directories to use for flight shuffle. Defaults to ["/tmp"].
        enable_ray_tracing: Enable tracing for Ray. Accessible in `/tmp/ray/session_latest/logs/daft` after the run completes. Defaults to False.
        scantask_splitting_level: How aggressively to split scan tasks. Setting this to `2` will use a more aggressive ScanTask splitting algorithm which might be more expensive to run but results in more even splits of partitions. Defaults to 1.
        native_parquet_writer: Whether to use the native parquet writer vs the pyarrow parquet writer. Defaults to `True`.
        use_experimental_distributed_engine: Whether to use the experimental distributed engine on the ray runner. Defaults to `True`.
            Note: Not all operations are currently supported, and daft will fallback to the current engine if necessary.
        min_cpu_per_task: Minimum CPU per task in the Ray runner. Defaults to 1.
    """
    # Replace values in the DaftExecutionConfig with user-specified overrides
    ctx = get_context()
    with ctx._lock:
        old_daft_execution_config = ctx._ctx._daft_execution_config if config is None else config

        new_daft_execution_config = old_daft_execution_config.with_config_values(
            scan_tasks_min_size_bytes=scan_tasks_min_size_bytes,
            scan_tasks_max_size_bytes=scan_tasks_max_size_bytes,
            max_sources_per_scan_task=max_sources_per_scan_task,
            broadcast_join_size_bytes_threshold=broadcast_join_size_bytes_threshold,
            parquet_split_row_groups_max_files=parquet_split_row_groups_max_files,
            sort_merge_join_sort_with_aligned_boundaries=sort_merge_join_sort_with_aligned_boundaries,
            hash_join_partition_size_leniency=hash_join_partition_size_leniency,
            sample_size_for_sort=sample_size_for_sort,
            num_preview_rows=num_preview_rows,
            parquet_target_filesize=parquet_target_filesize,
            parquet_target_row_group_size=parquet_target_row_group_size,
            parquet_inflation_factor=parquet_inflation_factor,
            csv_target_filesize=csv_target_filesize,
            csv_inflation_factor=csv_inflation_factor,
            shuffle_aggregation_default_partitions=shuffle_aggregation_default_partitions,
            partial_aggregation_threshold=partial_aggregation_threshold,
            high_cardinality_aggregation_threshold=high_cardinality_aggregation_threshold,
            read_sql_partition_size_bytes=read_sql_partition_size_bytes,
            enable_aqe=enable_aqe,
            default_morsel_size=default_morsel_size,
            shuffle_algorithm=shuffle_algorithm,
            flight_shuffle_dirs=flight_shuffle_dirs,
            pre_shuffle_merge_threshold=pre_shuffle_merge_threshold,
            enable_ray_tracing=enable_ray_tracing,
            scantask_splitting_level=scantask_splitting_level,
            native_parquet_writer=native_parquet_writer,
            use_experimental_distributed_engine=use_experimental_distributed_engine,
            min_cpu_per_task=min_cpu_per_task,
        )

        ctx._ctx._daft_execution_config = new_daft_execution_config
        return ctx



from __future__ import annotations

import contextlib
import datetime
import decimal
from pathlib import Path
from unittest.mock import patch

import pyarrow as pa
import pytest

import daft
from daft.io.object_store_options import io_config_to_storage_options
from daft.logical.schema import Schema
from tests.conftest import get_tests_daft_runner_name

PYARROW_LOWER_BOUND_SKIP = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) < (9, 0, 0)
pytestmark = pytest.mark.skipif(
    PYARROW_LOWER_BOUND_SKIP,
    reason="deltalake not supported on older versions of pyarrow",
)


class _FakeCommitProperties:
    def __init__(self, custom_metadata):
        self.custom_metadata = custom_metadata


@pytest.fixture
def custom_metadata():
    return {"CUSTOM_METADATA": "1"}


@pytest.fixture()
def commit_properties():
    @contextlib.contextmanager
    def _(deltalake):
        setattr(deltalake, "CommitProperties", _FakeCommitProperties)
        try:
            yield
        finally:
            delattr(deltalake, "CommitProperties")

    return _


def test_deltalake_write_basic(tmp_path, base_table):
    deltalake = pytest.importorskip("deltalake")
    path = tmp_path / "some_table"
    df = daft.from_arrow(base_table)
    result = df.write_deltalake(str(path))
    result = result.to_pydict()
    assert result["operation"] == ["ADD"]
    assert result["rows"] == [base_table.num_rows]

    read_delta = deltalake.DeltaTable(str(path))
    expected_schema = Schema.from_pyarrow_schema(pa.schema(read_delta.schema().to_arrow()))
    assert df.schema() == expected_schema
    assert read_delta.to_pyarrow_table() == base_table


def test_deltalake_multi_write_basic(tmp_path, base_table):
    deltalake = pytest.importorskip("deltalake")
    path = tmp_path / "some_table"
    df = daft.from_arrow(base_table)
    df.write_deltalake(str(path))

    result = df.write_deltalake(str(path))
    result = result.to_pydict()
    assert result["operation"] == ["ADD"]
    assert result["rows"] == [base_table.num_rows]

    read_delta = deltalake.DeltaTable(str(path))
    expected_schema = Schema.from_pyarrow_schema(pa.schema(read_delta.schema().to_arrow()))
    assert df.schema() == expected_schema
    assert read_delta.version() == 1
    assert read_delta.to_pyarrow_table() == pa.concat_tables([base_table, base_table])


def test_deltalake_write_cloud(base_table, cloud_paths):
    deltalake = pytest.importorskip("deltalake")
    path, io_config, _ = cloud_paths
    df = daft.from_arrow(base_table)
    result = df.write_deltalake(str(path), io_config=io_config)
    result = result.to_pydict()
    assert result["operation"] == ["ADD"]
    assert result["rows"] == [base_table.num_rows]
    storage_options = io_config_to_storage_options(io_config, path) if io_config is not None else None
    read_delta = deltalake.DeltaTable(str(path), storage_options=storage_options)
    expected_schema = Schema.from_pyarrow_schema(pa.schema(read_delta.schema().to_arrow()))
    assert df.schema() == expected_schema
    assert read_delta.to_pyarrow_table() == base_table


def test_deltalake_write_overwrite_basic(tmp_path):
    deltalake = pytest.importorskip("deltalake")
    path = tmp_path / "some_table"
    df1 = daft.from_pydict({"a": [1, 2]})
    df1.write_deltalake(str(path))

    df2 = daft.from_pydict({"a": [3, 4]})
    result = df2.write_deltalake(str(path), mode="overwrite")
    result = result.to_pydict()
    assert result["operation"] == ["ADD", "DELETE"]
    assert result["rows"] == [2, 2]

    read_delta = deltalake.DeltaTable(str(path))
    expected_schema = Schema.from_pyarrow_schema(pa.schema(read_delta.schema().to_arrow()))
    assert df2.schema() == expected_schema
    assert read_delta.to_pyarrow_table() == df2.to_arrow()


def test_deltalake_write_overwrite_cloud(cloud_paths):
    deltalake = pytest.importorskip("deltalake")
    path, io_config, _ = cloud_paths
    df1 = daft.from_pydict({"a": [1, 2]})
    df1.write_deltalake(str(path), io_config=io_config)

    df2 = daft.from_pydict({"a": [3, 4]})
    result = df2.write_deltalake(str(path), mode="overwrite", io_config=io_config)
    result = result.to_pydict()
    assert result["operation"] == ["ADD", "DELETE"]
    assert result["rows"] == [2, 2]

    storage_options = io_config_to_storage_options(io_config, path) if io_config is not None else None
    read_delta = deltalake.DeltaTable(str(path), storage_options=storage_options)
    expected_schema = Schema.from_pyarrow_schema(pa.schema(read_delta.schema().to_arrow()))
    assert df2.schema() == expected_schema
    assert read_delta.to_pyarrow_table() == df2.to_arrow()


@pytest.mark.skipif(
    get_tests_daft_runner_name() == "native",
    reason="Native executor does not support repartitioning",
)
def test_deltalake_write_overwrite_multi_partition(tmp_path):
    deltalake = pytest.importorskip("deltalake")
    path = tmp_path / "some_table"
    df1 = daft.from_pydict({"a": [1, 2, 3, 4]})
    df1 = df1.repartition(2)
    df1.write_deltalake(str(path))

    df2 = daft.from_pydict({"a": [5, 6, 7, 8]})
    df2 = df2.repartition(2)
    result = df2.write_deltalake(str(path), mode="overwrite")
    result = result.to_pydict()
    assert result["operation"] == ["ADD", "ADD", "DELETE", "DELETE"]

    read_delta = deltalake.DeltaTable(str(path))
    expected_schema = Schema.from_pyarrow_schema(pa.schema(read_delta.schema().to_arrow()))
    assert df2.schema() == expected_schema
    assert read_delta.to_pyarrow_table() == df2.to_arrow()


def test_deltalake_write_overwrite_schema(tmp_path):
    deltalake = pytest.importorskip("deltalake")
    path = tmp_path / "some_table"
    df1 = daft.from_pydict({"a": [1, 2]})
    df1.write_deltalake(str(path))

    df2 = daft.from_pydict({"b": [3, 4]})
    result = df2.write_deltalake(str(path), mode="overwrite", schema_mode="overwrite")
    result = result.to_pydict()
    assert result["operation"] == ["ADD", "DELETE"]

    read_delta = deltalake.DeltaTable(str(path))
    expected_schema = Schema.from_pyarrow_schema(pa.schema(read_delta.schema().to_arrow()))
    assert df2.schema() == expected_schema
    assert read_delta.to_pyarrow_table() == df2.to_arrow()


def test_deltalake_write_overwrite_error_schema(tmp_path):
    path = tmp_path / "some_table"
    df1 = daft.from_pydict({"a": [1, 2]})
    df1.write_deltalake(str(path), mode="overwrite")
    df2 = daft.from_pydict({"b": [3, 4]})
    with pytest.raises(ValueError):
        df2.write_deltalake(str(path), mode="overwrite")


def test_deltalake_write_error(tmp_path, base_table):
    path = tmp_path / "some_table"
    df = daft.from_arrow(base_table)
    df.write_deltalake(str(path), mode="error")
    with pytest.raises(AssertionError):
        df.write_deltalake(str(path), mode="error")


def test_deltalake_write_ignore(tmp_path):
    deltalake = pytest.importorskip("deltalake")
    path = tmp_path / "some_table"
    df1 = daft.from_pydict({"a": [1, 2]})
    df1.write_deltalake(str(path), mode="ignore")
    df2 = daft.from_pydict({"a": [3, 4]})
    result = df2.write_deltalake(str(path), mode="ignore")
    result = result.to_arrow()
    assert result.num_rows == 0

    read_delta = deltalake.DeltaTable(str(path))
    expected_schema = Schema.from_pyarrow_schema(pa.schema(read_delta.schema().to_arrow()))
    assert df1.schema() == expected_schema
    assert read_delta.to_pyarrow_table() == df1.to_arrow()


@pytest.mark.skipif(
    get_tests_daft_runner_name() == "native",
    reason="Native executor does not support repartitioning",
)
def test_deltalake_write_with_empty_partition(tmp_path, base_table):
    deltalake = pytest.importorskip("deltalake")
    path = tmp_path / "some_table"
    df = daft.from_arrow(base_table).into_partitions(4)
    result = df.write_deltalake(str(path))
    result = result.to_pydict()
    assert result["operation"] == ["ADD", "ADD", "ADD"]
    assert result["rows"] == [1, 1, 1]

    read_delta = deltalake.DeltaTable(str(path))
    expected_schema = Schema.from_pyarrow_schema(pa.schema(read_delta.schema().to_arrow()))
    assert df.schema() == expected_schema
    assert read_delta.to_pyarrow_table() == base_table


def check_equal_both_daft_and_delta_rs(df: daft.DataFrame, path: Path, sort_order: list[tuple[str, str]]):
    deltalake = pytest.importorskip("deltalake")

    arrow_df = df.to_arrow().sort_by(sort_order)

    read_daft = daft.read_deltalake(str(path))
    assert read_daft.schema() == df.schema()
    assert read_daft.to_arrow().sort_by(sort_order) == arrow_df

    read_delta = deltalake.DeltaTable(str(path))
    expected_schema = Schema.from_pyarrow_schema(pa.schema(read_delta.schema().to_arrow()))
    assert df.schema() == expected_schema
    assert read_delta.to_pyarrow_table().cast(expected_schema.to_pyarrow_schema()).sort_by(sort_order) == arrow_df


@pytest.mark.parametrize(
    "partition_cols,num_partitions",
    [
        (["int"], 3),
        (["float"], 3),
        (["str"], 3),
        pytest.param(["bin"], 3, marks=pytest.mark.xfail(reason="Binary partitioning is not yet supported")),
        (["bool"], 3),
        (["datetime"], 3),
        (["date"], 3),
        (["decimal"], 3),
        (["int", "float"], 4),
    ],
)
def test_deltalake_write_partitioned(tmp_path, partition_cols, num_partitions):
    path = tmp_path / "some_table"
    df = daft.from_pydict(
        {
            "int": [1, 1, 2, None],
            "float": [1.1, 2.2, 2.2, None],
            "str": ["foo", "foo", "bar", None],
            "bin": [b"foo", b"foo", b"bar", None],
            "bool": [True, True, False, None],
            "datetime": [
                datetime.datetime(2024, 2, 10),
                datetime.datetime(2024, 2, 10),
                datetime.datetime(2024, 2, 11),
                None,
            ],
            "date": [datetime.date(2024, 2, 10), datetime.date(2024, 2, 10), datetime.date(2024, 2, 11), None],
            "decimal": pa.array(
                [decimal.Decimal("1111.111"), decimal.Decimal("1111.111"), decimal.Decimal("2222.222"), None],
                type=pa.decimal128(7, 3),
            ),
        }
    )
    result = df.write_deltalake(str(path), partition_cols=partition_cols)
    result = result.to_pydict()
    assert len(result["operation"]) == num_partitions
    assert all(op == "ADD" for op in result["operation"])
    assert sum(result["rows"]) == len(df)

    sort_order = [("int", "ascending"), ("float", "ascending")]
    check_equal_both_daft_and_delta_rs(df, path, sort_order)


def test_deltalake_write_partitioned_empty(tmp_path):
    path = tmp_path / "some_table"

    df = daft.from_arrow(pa.schema([("int", pa.int64()), ("string", pa.string())]).empty_table())

    df.write_deltalake(str(path), partition_cols=["int"])

    check_equal_both_daft_and_delta_rs(df, path, [("int", "ascending")])


def test_deltalake_write_partitioned_some_empty(tmp_path):
    path = tmp_path / "some_table"

    df = daft.from_pydict({"int": [1, 2, 3, None], "string": ["foo", "foo", "bar", None]}).into_partitions(5)

    df.write_deltalake(str(path), partition_cols=["int"])

    check_equal_both_daft_and_delta_rs(df, path, [("int", "ascending")])


def test_deltalake_write_partitioned_existing_table(tmp_path):
    path = tmp_path / "some_table"

    df1 = daft.from_pydict({"int": [1], "string": ["foo"]})
    result = df1.write_deltalake(str(path), partition_cols=["int"])
    result = result.to_pydict()
    assert result["operation"] == ["ADD"]
    assert result["rows"] == [1]

    df2 = daft.from_pydict({"int": [1, 2], "string": ["bar", "bar"]})
    with pytest.raises(ValueError):
        df2.write_deltalake(str(path), partition_cols=["string"])

    result = df2.write_deltalake(str(path))
    result = result.to_pydict()
    assert result["operation"] == ["ADD", "ADD"]
    assert result["rows"] == [1, 1]

    check_equal_both_daft_and_delta_rs(df1.concat(df2), path, [("int", "ascending"), ("string", "ascending")])


def test_deltalake_write_roundtrip(tmp_path):
    path = tmp_path / "some_table"
    df = daft.from_pydict({"a": [1, 2, 3, 4]})
    df.write_deltalake(str(path))

    read_df = daft.read_deltalake(str(path))
    assert df.schema() == read_df.schema()
    assert df.to_arrow() == read_df.to_arrow()


def test_custom_metadata_added_for_new_table(tmp_path, custom_metadata):
    # import deltalake
    deltalake = pytest.importorskip("deltalake")

    path = tmp_path / "some_table"
    df = daft.from_pydict({"a": [1, 2, 3, 4]})
    df.write_deltalake(str(path), custom_metadata=custom_metadata)

    table = deltalake.DeltaTable(path)
    history = table.history(1)

    assert custom_metadata.items() <= history[0].items()


def test_custom_metadata_updated_for_existing_table(tmp_path, custom_metadata):
    """Tests for deltalake version installed in the current environment (currently 0.19.2)."""
    # import deltalake
    deltalake = pytest.importorskip("deltalake")

    path = tmp_path / "some_table"
    df = daft.from_pydict({"a": [1, 2, 3, 4]})
    df.write_deltalake(str(path))

    df = daft.from_pydict({"a": [5, 6]})
    df.write_deltalake(str(path), custom_metadata=custom_metadata, mode="append")

    table = deltalake.DeltaTable(path)
    history = table.history(1)

    assert custom_metadata.items() <= history[0].items()


def test_custom_metadata_updated_for_existing_table_with_commit_properties(
    tmp_path, custom_metadata, commit_properties
):
    deltalake = pytest.importorskip("deltalake")
    from deltalake._internal import RawDeltaTable

    # write once to get into the table is not None path
    path = tmp_path / "some_table"
    df = daft.from_pydict({"a": [1, 2, 3, 4]})
    df.write_deltalake(str(path))

    # Add mocked CommitProperties class introduced in 0.20.0
    with commit_properties(deltalake), patch.object(RawDeltaTable, "create_write_transaction") as mock_method:
        df = daft.from_pydict({"a": [5, 6]})
        df.write_deltalake(str(path), custom_metadata=custom_metadata, mode="append")

        mock_method.assert_called_once()
        (_, _, _, _, _, custom_metadata_arg) = mock_method.call_args[0]

        assert isinstance(custom_metadata_arg, _FakeCommitProperties)
        assert custom_metadata_arg.custom_metadata == custom_metadata



from __future__ import annotations

import pytest

from daft.expressions import col, lit
from daft.recordbatch import MicroPartition

ENDSWITH_DATA = ["x_foo", "y_foo", "z_bar"]
STARTSWITH_DATA = ["foo_x", "foo_y", "bar_z"]
CONTAINS_DATA = ["x_foo_x", "y_foo_y", "z_bar_z"]


@pytest.mark.parametrize(
    ["expr", "data"],
    [
        (col("col").str.endswith("foo"), ENDSWITH_DATA),
        (col("col").str.endswith(lit("foo")), ENDSWITH_DATA),
        (col("col").str.endswith(col("emptystrings") + lit("foo")), ENDSWITH_DATA),
        (col("col").str.startswith("foo"), STARTSWITH_DATA),
        (col("col").str.startswith(lit("foo")), STARTSWITH_DATA),
        (col("col").str.startswith(col("emptystrings") + lit("foo")), STARTSWITH_DATA),
        (col("col").str.contains("foo"), CONTAINS_DATA),
        (col("col").str.contains(lit("foo")), CONTAINS_DATA),
        (col("col").str.contains(col("emptystrings") + lit("foo")), CONTAINS_DATA),
    ],
)
def test_utf8_substrs(expr, data):
    table = MicroPartition.from_pydict({"col": data, "emptystrings": ["", "", ""]})
    result = table.eval_expression_list([expr])
    assert result.to_pydict() == {"col": [True, True, False]}



from __future__ import annotations

import argparse
import logging
import typing
from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path

import duckdb
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import ray

import daft
from benchmarking.tpcds.datagen import gen_tpcds
from benchmarking.tpcds.helpers import parse_questions_str

from ..tpch import __main__ as tpch
from ..tpch import ray_job_runner
from . import helpers

logger = logging.getLogger(__name__)

SQL_QUERIES_PATH = Path(__file__).parent / "queries"


@dataclass
class ParsedArgs:
    tpcds_gen_folder: Path
    scale_factor: float
    questions: str
    ray_address: str | None
    dry_run: bool
    validate: bool
    cast_decimals: bool
    save_csv_on_failure: bool


@dataclass
class RunArgs:
    scaled_tpcds_gen_folder: Path
    query_indices: list[int]
    ray_address: str | None
    dry_run: bool
    validate: bool
    save_csv_on_failure: bool


@dataclass
class Result:
    index: int
    duration: timedelta | None
    error_msg: str | None
    is_correct: bool | None = None
    validation_error: str | None = None

    def __repr__(self) -> str:
        if self.duration and self.error_msg:
            typing.assert_never("Both duration and error_msg are not None")
        elif self.duration:
            validation_str = ""
            if self.is_correct is not None:
                validation_str = f" - Correct: {self.is_correct}"
                if not self.is_correct and self.validation_error:
                    validation_str += f" - Validation Error: {self.validation_error}"
            return f"(Q{self.index} SUCCESS - duration: {self.duration}{validation_str})"
        elif self.error_msg:
            return f"(Q{self.index} FAILURE - error msg: {self.error_msg})"
        else:
            typing.assert_never("Both duration and error_msg are None")


def setup_duckdb_catalog(data_dir: Path) -> duckdb.DuckDBPyConnection:
    """Setup DuckDB connection using the existing database file."""
    return duckdb.connect(str(data_dir / "tpcds.db"))


def _is_string(tp: pa.DataType) -> bool:
    return pa.types.is_string(tp) or pa.types.is_large_string(tp)


def _is_numeric(tp: pa.DataType) -> bool:
    return pa.types.is_floating(tp) or pa.types.is_decimal(tp) or pa.types.is_integer(tp)


def _types_compatible(t1: pa.DataType, t2: pa.DataType) -> bool:
    """Decide whether two Arrow data types should be considered the same."""
    if t1.equals(t2):
        return True

    if _is_string(t1) and _is_string(t2):
        return True

    if _is_numeric(t1) and _is_numeric(t2):
        return True

    return False


def _arrays_equal(
    lhs: pa.ChunkedArray,
    rhs: pa.ChunkedArray,
    *,
    rtol: float,
    atol: float,
) -> bool:
    """Return True if two ChunkedArrays are equal within tolerance."""
    lhs = lhs.combine_chunks()
    rhs = rhs.combine_chunks()

    if pc.any(pc.xor(pc.is_null(lhs), pc.is_null(rhs))).as_py():
        return False

    t1, t2 = lhs.type, rhs.type

    if _is_numeric(t1) and _is_numeric(t2):
        lhs_f = pc.cast(lhs, pa.float64(), safe=False)
        rhs_f = pc.cast(rhs, pa.float64(), safe=False)

        mask = pc.invert(pc.is_null(lhs_f))
        lhs_np = lhs_f.filter(mask).to_numpy()
        rhs_np = rhs_f.filter(mask).to_numpy()

        return np.allclose(lhs_np, rhs_np, rtol=rtol, atol=atol)

    if _is_string(t1) and _is_string(t2):
        lhs_casted = pc.cast(lhs, pa.large_string(), safe=False)
        rhs_casted = pc.cast(rhs, pa.large_string(), safe=False)
        return lhs_casted.equals(rhs_casted)

    return lhs.equals(rhs)


def validate_query_results(
    daft_results,
    query: str,
    data_dir: Path,
    *,
    rtol: float = 1e-3,
    atol: float = 1e-3,
    save_csv_on_failure: bool = False,
    query_index: int | None = None,
) -> tuple[bool, str | None]:
    """Compare Daft and DuckDB results."""
    try:
        conn = setup_duckdb_catalog(data_dir)
        daft_tbl = daft_results.to_arrow()
        duck_tbl = conn.execute(query).arrow()

        daft_csv_name = f"daft_q{query_index}_results.csv" if query_index is not None else "daft_results.csv"
        duckdb_csv_name = f"duckdb_q{query_index}_results.csv" if query_index is not None else "duckdb_results.csv"

        csv_saved_message = f" Results saved to {daft_csv_name} and {duckdb_csv_name}." if save_csv_on_failure else ""

        if daft_tbl.num_rows != duck_tbl.num_rows:
            if save_csv_on_failure:
                daft_tbl.to_pandas().to_csv(daft_csv_name, index=False)
                duck_tbl.to_pandas().to_csv(duckdb_csv_name, index=False)
            return (
                False,
                f"Row counts differ: Daft has {daft_tbl.num_rows} rows, DuckDB has {duck_tbl.num_rows} rows.{csv_saved_message}",
            )

        if daft_tbl.schema.names != duck_tbl.schema.names:
            if save_csv_on_failure:
                daft_tbl.to_pandas().to_csv(daft_csv_name, index=False)
                duck_tbl.to_pandas().to_csv(duckdb_csv_name, index=False)
            return (
                False,
                f"Column name order differs.\\nDaft column names: {daft_tbl.schema.names}\\nDuckDB column names: {duck_tbl.schema.names}.{csv_saved_message}",
            )

        for field_daft, field_duck in zip(daft_tbl.schema, duck_tbl.schema):
            name = field_daft.name
            type_daft = field_daft.type
            type_duck = field_duck.type

            if not _types_compatible(type_daft, type_duck):
                if save_csv_on_failure:
                    daft_tbl.to_pandas().to_csv(daft_csv_name, index=False)
                    duck_tbl.to_pandas().to_csv(duckdb_csv_name, index=False)
                return (
                    False,
                    f"Incompatible types for column '{name}': Daft type is {type_daft}, DuckDB type is {type_duck}.{csv_saved_message}",
                )

            lhs_arr = daft_tbl.column(name)
            rhs_arr = duck_tbl.column(name)

            if not _arrays_equal(lhs_arr, rhs_arr, rtol=rtol, atol=atol):
                if save_csv_on_failure:
                    daft_tbl.to_pandas().to_csv(daft_csv_name, index=False)
                    duck_tbl.to_pandas().to_csv(duckdb_csv_name, index=False)
                if _is_string(type_daft) and _is_string(type_duck):
                    return (
                        False,
                        f"Column '{name}' (types: Daft={type_daft}, DuckDB={type_duck}) differs.{csv_saved_message}",
                    )
                else:
                    return (
                        False,
                        f"Column '{name}' (types: Daft={type_daft}, DuckDB={type_duck}) differs beyond tolerance "
                        f"(rtol={rtol}, atol={atol}).{csv_saved_message}",
                    )

        return True, None

    except Exception as e:
        return False, f"Validation error during comparison process: {e!s}"


def run_query_on_ray(
    run_args: RunArgs,
) -> list[Result]:
    ray.init(address=run_args.ray_address if run_args.ray_address else None)
    results = []

    for query_index in run_args.query_indices:
        working_dir = Path("benchmarking") / "tpcds"
        ray_entrypoint_script = "ray_entrypoint.py"
        duration = None
        error_msg = None
        is_correct = None
        validation_error = None
        try:
            start = datetime.now()
            ray_job_runner.run_on_ray(
                run_args.ray_address,
                {
                    "entrypoint": f"python {ray_entrypoint_script} --tpcds-gen-folder 'data/0.01' --question {query_index} {'--dry-run' if run_args.dry_run else ''}",
                    "runtime_env": {
                        "working_dir": working_dir,
                    },
                },
            )
            end = datetime.now()
            duration = end - start
        except Exception as e:
            error_msg = str(e)

        results.append(
            Result(
                index=query_index,
                duration=duration,
                error_msg=error_msg,
                is_correct=is_correct,
                validation_error=validation_error,
            )
        )

    return results


def run_query_on_local(
    run_args: RunArgs,
) -> list[Result]:
    catalog = helpers.generate_catalog(run_args.scaled_tpcds_gen_folder)
    results = []

    for query_index in run_args.query_indices:
        query_file = SQL_QUERIES_PATH / f"{query_index:02}.sql"
        with open(query_file) as f:
            query = f.read()

        start = datetime.now()

        duration = None
        error_msg = None
        is_correct = None
        validation_error = None

        try:
            daft.sql(query, catalog=catalog).explain(show_all=True)
            if not run_args.dry_run:
                daft_results = daft.sql(query, catalog=catalog).collect()

                if run_args.validate and not run_args.dry_run:
                    is_correct, validation_error = validate_query_results(
                        daft_results,
                        query,
                        run_args.scaled_tpcds_gen_folder,
                        save_csv_on_failure=run_args.save_csv_on_failure,
                        query_index=query_index,
                    )
                    if is_correct:
                        logger.info("Query %s results validated successfully", query_index)
                    else:
                        logger.warning("Query %s validation failed: %s", query_index, validation_error)

            end = datetime.now()
            duration = end - start
        except Exception as e:
            error_msg = str(e)

        results.append(
            Result(
                index=query_index,
                duration=duration,
                error_msg=error_msg,
                is_correct=is_correct,
                validation_error=validation_error,
            )
        )

    return results


def run_benchmarks(
    run_args: RunArgs,
) -> list[Result]:
    logger.info(
        "Running the following questions: %s",
        run_args.query_indices,
    )

    runner = tpch.get_daft_benchmark_runner_name()

    logger.info(
        "Running on the following runner: %s",
        runner,
    )

    if runner == "ray":
        return run_query_on_ray(run_args)
    elif runner == "native":
        return run_query_on_local(run_args)
    else:
        typing.assert_never(runner)


def main(args: ParsedArgs):
    scaled_tpcds_gen_folder = args.tpcds_gen_folder / str(args.scale_factor)
    gen_tpcds(scaled_tpcds_gen_folder, args.scale_factor, cast_decimal=args.cast_decimals)
    query_indices = parse_questions_str(args.questions)
    results = run_benchmarks(
        RunArgs(
            scaled_tpcds_gen_folder=scaled_tpcds_gen_folder,
            query_indices=query_indices,
            ray_address=args.ray_address,
            dry_run=args.dry_run,
            validate=args.validate,
            save_csv_on_failure=args.save_csv_on_failure,
        )
    )

    # TODO(ronnie): improve visualization of results; simply printing them to console is not the best way...
    print(f"{results=}")


if __name__ == "__main__":
    logging.basicConfig(level="INFO")

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--tpcds-gen-folder",
        default="benchmarking/tpcds/data",
        type=Path,
        help="Path to the folder containing the TPC-DS dsdgen tool and generated data",
    )
    parser.add_argument("--scale-factor", default=0.01, type=float, help="Scale factor to run on in GB")
    parser.add_argument("--questions", default="*", type=str, help="The questions to run")
    parser.add_argument("--ray-address", type=str, help="The address of the head node of the ray cluster")
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Whether to run in dry-run mode; if true, only the plan will be printed, but no query will be executed",
    )
    parser.add_argument(
        "--cast-decimals",
        action="store_true",
        help="Cast decimal columns to float64 before running queries",
    )
    parser.add_argument(
        "--validate",
        action="store_true",
        help="Validate query results against DuckDB reference implementation",
    )
    parser.add_argument(
        "--save-csv-on-failure",
        action="store_true",
        dest="save_csv_on_failure",
        help="Save Daft and DuckDB results to CSV if validation fails (default: False)",
    )
    args = parser.parse_args()

    tpcds_gen_folder: Path = args.tpcds_gen_folder
    assert args.scale_factor > 0

    main(
        ParsedArgs(
            tpcds_gen_folder=tpcds_gen_folder,
            scale_factor=args.scale_factor,
            questions=args.questions,
            ray_address=args.ray_address,
            dry_run=args.dry_run,
            validate=args.validate,
            cast_decimals=args.cast_decimals,
            save_csv_on_failure=args.save_csv_on_failure,
        )
    )



from __future__ import annotations

import copy

import pyarrow as pa
import pytest

from daft import col
from daft.logical.schema import Schema
from daft.recordbatch.micropartition import MicroPartition


@pytest.mark.parametrize(
    "mp",
    [
        MicroPartition.from_pydict({"a": pa.array([1, 2, 3], type=pa.int64())}),
        MicroPartition.from_pydict({"a": pa.array([], type=pa.int64())}),  # 1 empty table
        MicroPartition.empty(Schema.from_pyarrow_schema(pa.schema({"a": pa.int64()}))),  # No tables
    ],
)
def test_pickling_loaded(mp) -> None:
    assert mp.to_arrow() == copy.deepcopy(mp).to_arrow()


def test_pickling_unloaded() -> None:
    mp = MicroPartition.read_parquet("tests/assets/parquet-data/parquet-with-schema-metadata.parquet")
    assert copy.deepcopy(mp).to_arrow() == mp.to_arrow()


@pytest.mark.parametrize(
    "mp",
    [
        MicroPartition.from_pydict(
            {"a": pa.array([1, 2, 3], type=pa.int64()), "b": pa.array([4.0, 5.0, 6.0], type=pa.float64())}
        ),
        MicroPartition.empty(Schema.from_pyarrow_schema(pa.schema({"a": pa.int64(), "b": pa.float64()}))),  # No tables
    ],
)
def test_pickling_duplicate_column_names(mp) -> None:
    # To force duplicate column names
    mp = mp.eval_expression_list([col("a"), col("b").alias("a")])

    assert mp.to_arrow() == copy.deepcopy(mp).to_arrow()



from __future__ import annotations

import pytest


def test_select_dataframe_missing_col(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)

    with pytest.raises(ValueError):
        df = df.select("foo", "sepal_length")


def test_select_dataframe(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)
    df = df.select("sepal_length", "sepal_width")
    assert df.column_names == ["sepal_length", "sepal_width"]


def test_multiple_select_same_col(make_df, valid_data: list[dict[str, float]]):
    df = make_df(valid_data)
    df = df.select(df["sepal_length"], df["sepal_length"].alias("sepal_length_2"))
    pdf = df.to_pandas()
    assert len(pdf.columns) == 2
    assert pdf.columns.to_list() == ["sepal_length", "sepal_length_2"]


def test_select_ordering(make_df, valid_data: list[dict[str, float]]):
    df = make_df(valid_data)
    df = df.select(
        df["variety"], df["petal_length"].alias("foo"), df["sepal_length"], df["sepal_width"], df["petal_width"]
    )
    df = df.collect()
    assert df.column_names == ["variety", "foo", "sepal_length", "sepal_width", "petal_width"]



from __future__ import annotations

import os
import warnings
from typing import TYPE_CHECKING, Any, Literal

import turbopuffer

from daft.datatype import DataType
from daft.dependencies import pc
from daft.expressions import ExpressionsProjection
from daft.io import DataSink
from daft.io.sink import WriteResult
from daft.recordbatch import MicroPartition
from daft.schema import Schema

if TYPE_CHECKING:
    from collections.abc import Iterator
    from types import ModuleType

    from daft.dependencies import pa
    from daft.expressions import Expression


class TurbopufferDataSink(DataSink[turbopuffer.types.NamespaceWriteResponse]):
    """WriteSink for writing data to a Turbopuffer namespace."""

    _namespace: str | ExpressionsProjection

    @staticmethod
    def _import_turbopuffer() -> ModuleType:
        try:
            import turbopuffer

            return turbopuffer
        except ImportError:
            raise ImportError(
                "turbopuffer is not installed. Please install turbopuffer using `pip install turbopuffer`"
            )

    @staticmethod
    def _check_namespace_name(namespace_name: str) -> None:
        if not isinstance(namespace_name, str):
            raise ValueError(f"Namespace name must be a string, got {namespace_name} with type {type(namespace_name)}")
        if not (1 <= len(namespace_name) <= 128):
            raise ValueError(
                f"Namespace name must be between 1 and 128 characters, got length {len(namespace_name)}. For more details, see: https://turbopuffer.com/docs/write"
            )
        if not all(c.isalnum() or c in "-_." for c in namespace_name):
            raise ValueError(
                f"Namespace name can only contain alphanumeric characters, hyphens, underscores and periods. Got: {namespace_name}. For more details, see: https://turbopuffer.com/docs/write"
            )
        return

    def __init__(
        self,
        namespace: str | Expression,
        api_key: str | None = None,
        region: str | None = None,
        distance_metric: Literal["cosine_distance", "euclidean_squared"] | None = None,
        schema: dict[str, Any] | None = None,
        id_column: str | None = None,
        vector_column: str | None = None,
        client_kwargs: dict[str, Any] | None = None,
        write_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the Turbopuffer data sink.

        This data sink transforms each row of the dataframe into a turbopuffer document.
        This means that an `id` column is always required. Optionally, the `id_column` parameter can be used to specify the column name to used for the id column.
        Note that the column with the name specified by `id_column` will be renamed to "id" when written to turbopuffer.

        A `vector` column is required if the namespace has a vector index. Optionally, the `vector_column` parameter can be used to specify the column name to used for the vector index.
        Note that the column with the name specified by `vector_column` will be renamed to "vector" when written to turbopuffer.

        All other columns become attributes.

        The namespace parameter can be either a string (for a single namespace) or an expression (for multiple namespaces).
        When using an expression, the data will be partitioned by the computed namespace values and written to each namespace separately.

        For more details on parameters, please see the turbopuffer documentation: https://turbopuffer.com/docs/write

        Args:
            namespace: The namespace to write to. Can be a string for a single namespace or an expression for multiple namespaces.
            api_key: Turbopuffer API key.
            region: Turbopuffer region.
            distance_metric: Distance metric for vector similarity ("cosine_distance", "euclidean_squared").
            schema: Optional manual schema specification.
            id_column: Optional column name for the id column. The data sink will automatically rename the column to "id" for the id column.
            vector_column: Optional column name for the vector index column. The data sink will automatically rename the column to "vector" for the vector index.
            client_kwargs: Optional dictionary of arguments to pass to the Turbopuffer client constructor.
                Explicit arguments (api_key, region) will be merged into client_kwargs.
            write_kwargs: Optional dictionary of arguments to pass to the namespace.write() method.
                Explicit arguments (distance_metric, schema) will be merged into write_kwargs.
        """
        if isinstance(namespace, str):
            TurbopufferDataSink._check_namespace_name(namespace)
            self._namespace = namespace
        else:
            self._namespace = ExpressionsProjection([namespace])

        self._api_key = api_key or os.getenv("TURBOPUFFER_API_KEY")
        if not self._api_key:
            warnings.warn("Turbopuffer API key is not set and TURBOPUFFER_API_KEY environment variable is not found")

        self._region = region
        self._distance_metric = distance_metric
        self._schema = schema or {}
        self._id_column = id_column
        self._vector_column = vector_column

        # Merge explicit arguments into kwargs.
        self._client_kwargs = client_kwargs or {}
        if api_key is not None:
            if "api_key" in self._client_kwargs:
                raise ValueError("api_key is already set in client_kwargs")
            self._client_kwargs["api_key"] = api_key
        if region is not None:
            if "region" in self._client_kwargs:
                raise ValueError("region is already set in client_kwargs")
            self._client_kwargs["region"] = region

        self._write_kwargs = write_kwargs or {}
        if distance_metric is not None:
            if "distance_metric" in self._write_kwargs:
                raise ValueError("distance_metric is already set in write_kwargs")
            self._write_kwargs["distance_metric"] = distance_metric
        if schema is not None:
            if "schema" in self._write_kwargs:
                raise ValueError("schema is already set in write_kwargs")
            self._write_kwargs["schema"] = schema

        self._result_schema = Schema._from_field_name_and_types([("write_responses", DataType.python())])

    def name(self) -> str:
        return "Turbopuffer Write"

    def schema(self) -> Schema:
        return self._result_schema

    def _prepare_arrow_table(self, arrow_table: pa.Table) -> pa.Table:
        if self._id_column:
            if self._id_column not in arrow_table.column_names:
                raise ValueError(f"ID column {self._id_column} not found in schema, cannot use as column for id")
            arrow_table = arrow_table.rename_columns({self._id_column: "id"})

        if self._vector_column:
            if self._vector_column not in arrow_table.column_names:
                raise ValueError(
                    f"Vector column {self._vector_column} not found in schema, cannot use as column for vector index"
                )
            arrow_table = arrow_table.rename_columns({self._vector_column: "vector"})

        return arrow_table.filter(~pc.field("id").is_null())

    def write(
        self, micropartitions: Iterator[MicroPartition]
    ) -> Iterator[WriteResult[turbopuffer.types.NamespaceWriteResponse]]:
        """Writes micropartitions to Turbopuffer namespace(s)."""
        turbopuffer = TurbopufferDataSink._import_turbopuffer()
        tpuf = turbopuffer.Turbopuffer(**self._client_kwargs)

        for micropartition in micropartitions:
            if isinstance(self._namespace, str):
                # Namespace is a string. Write all data to this namespace.
                arrow_table = self._prepare_arrow_table(micropartition.to_arrow())

                namespace = tpuf.namespace(self._namespace)
                write_response = namespace.write(
                    upsert_rows=arrow_table.to_pylist(),
                    **self._write_kwargs,
                )

                yield WriteResult(
                    result=write_response,
                    bytes_written=arrow_table.nbytes,
                    rows_written=arrow_table.num_rows,
                )
            else:
                # Namespace is an expression. Partition the data by namespace then write to each namespace.
                (partitioned_data, partition_keys) = micropartition.partition_by_value(self._namespace)
                for data, namespace_name in zip(partitioned_data, partition_keys.get_column(0)):
                    if len(data) == 0:
                        continue

                    arrow_table = self._prepare_arrow_table(data.to_arrow())

                    TurbopufferDataSink._check_namespace_name(namespace_name)
                    namespace = tpuf.namespace(namespace_name)
                    write_response = namespace.write(
                        upsert_rows=arrow_table.to_pylist(),
                        distance_metric=self._distance_metric,
                        schema=self._schema,
                    )

                    yield WriteResult(
                        result=write_response,
                        bytes_written=arrow_table.nbytes,
                        rows_written=arrow_table.num_rows,
                    )

    def finalize(self, write_results: list[WriteResult[turbopuffer.types.NamespaceWriteResponse]]) -> MicroPartition:
        """Finalizes the write process and returns summary statistics."""
        result_table = MicroPartition.from_pydict(
            {
                "write_responses": write_results,
            }
        )

        return result_table



from __future__ import annotations

import itertools

import pytest

import daft
from daft.datatype import DataType
from daft.expressions import col, lit


def test_repr_cosine():
    a = col("a")
    b = col("b")
    y = a.embedding.cosine_distance(b)
    repr_out = repr(y)
    assert repr_out == "cosine_distance(col(a), col(b))"


def _generate_valid_dtype_pairs_for_cosine_distance():
    valid_inner_dtypes = [DataType.float64(), DataType.float32(), DataType.int8()]
    pairs = []

    for inner_dtype in valid_inner_dtypes:
        outer_dtypes = [
            DataType.fixed_size_list(inner_dtype, 3),
            DataType.embedding(inner_dtype, 3),
        ]
        pairs.extend(itertools.product(outer_dtypes, outer_dtypes))

    return pairs


valid_cosine_distance_dtype_pairs = _generate_valid_dtype_pairs_for_cosine_distance()


@pytest.mark.parametrize("dtype_pair", valid_cosine_distance_dtype_pairs)
def test_cosine_floats(dtype_pair):
    dtype1, dtype2 = dtype_pair
    data = {
        "source": [[1.0, 1.11, 1.01], [2, 2, 2], [3, 3, 3]],
    }
    query = [1.0, 1.11, 1.01]

    df = daft.from_pydict(data)
    df = df.with_column("source", col("source").cast(dtype1))
    df = df.with_column("query", lit(query).cast(dtype2))

    res = df.select(col("source").embedding.cosine_distance(col("query"))).to_pydict()

    def cosine_dist_brute_force(x, y):
        import math

        xy = sum(xi * yi for xi, yi in zip(x, y))
        x_sq = math.sqrt(sum(xi**2 for xi in x))
        y_sq = math.sqrt(sum(yi**2 for yi in y))
        return 1.0 - xy / (x_sq * y_sq)

    # If dtype is int8, cast data and query to int
    if dtype1.dtype.is_int8():
        expected = [cosine_dist_brute_force(list(map(int, query)), list(map(int, x))) for x in data["source"]]
    else:
        expected = [cosine_dist_brute_force(query, x) for x in data["source"]]

    # check if they are approximately equal
    for a, b in zip(res["source"], expected):
        assert pytest.approx(b) == pytest.approx(a, abs=1e-5)


@pytest.mark.parametrize("dtype_pair", valid_cosine_distance_dtype_pairs)
def test_pairwise_cosine_distance(dtype_pair):
    dtype1, dtype2 = dtype_pair
    data = {
        "e1": [[1, 2, 3], [1, 2, 3]],
        "e2": [[1, 2, 3], [-1, -2, -3]],
    }
    df = daft.from_pydict(data)
    res = df.with_column(
        "distance",
        df["e1"].cast(dtype1).embedding.cosine_distance(df["e2"].cast(dtype2)),
    ).to_pydict()

    assert res["distance"] == [0.0, 2.0]


@pytest.mark.parametrize(
    "data",
    [
        # Non fixed size list
        [[1, 2, 3], [1, 2, 3]],
        # integers
        [1, 2, 3],
        # floats
        [1.0, 2.0, 3.0],
        # strings
        ["1", "2", "3"],
    ],
)
def test_cosine_distance_dtype_not_fixed_size(data):
    data = {
        "source": data,
    }
    df = daft.from_pydict(data)
    with pytest.raises(
        ValueError,
        match="Expected inputs to 'cosine_distance' to be fixed size list or embedding",
    ):
        df.with_column("distance", df["source"].embedding.cosine_distance(df["source"]))


def test_cosine_distance_dtype_size_mismatch():
    data = {
        "source": [[1, 2, 3], [1, 2, 3]],
        "query": [[1, 2, 3, 4], [1, 2, 3, 4]],
    }

    df = daft.from_pydict(data)
    df = df.with_column("source", df["source"].cast(DataType.embedding(DataType.float32(), 3)))
    df = df.with_column("query", df["query"].cast(DataType.embedding(DataType.float32(), 4)))

    with pytest.raises(ValueError, match="Expected inputs to 'cosine_distance' to have the same size"):
        df.with_column("distance", df["source"].embedding.cosine_distance(df["query"]))


def test_cosine_distance_dtype_precision_mismatch():
    data = {
        "source": [[1, 2, 3], [1, 2, 3]],
        "query": [[1, 2, 3], [1, 2, 3]],
    }

    df = daft.from_pydict(data)
    df = df.with_column("source", df["source"].cast(DataType.embedding(DataType.float64(), 3)))
    df = df.with_column("query", df["query"].cast(DataType.embedding(DataType.float32(), 3)))

    with pytest.raises(
        ValueError,
        match="Expected inputs to 'cosine_distance' to have the same inner dtype",
    ):
        df.with_column("distance", df["source"].embedding.cosine_distance(df["query"]))



from __future__ import annotations

import daft.daft as native
from daft.expressions import Expression, col, lit


def monotonically_increasing_id() -> Expression:
    """Generates a column of monotonically increasing unique ids.

    The implementation puts the partition number in the upper 28 bits, and the row number in each partition
    in the lower 36 bits. This allows for 2^28 ≈ 268 million partitions and 2^36 ≈ 68 billion rows per partition.

    Returns:
        Expression: An expression that generates monotonically increasing IDs

    Examples:
        >>> import daft
        >>> from daft.functions import monotonically_increasing_id
        >>> daft.context.set_runner_ray()  # doctest: +SKIP
        >>>
        >>> df = daft.from_pydict({"a": [1, 2, 3, 4]}).into_partitions(2)
        >>> df = df.with_column("id", monotonically_increasing_id())
        >>> df.show()  # doctest: +SKIP
        ╭───────┬─────────────╮
        │ a     ┆ id          │
        │ ---   ┆ ---         │
        │ Int64 ┆ UInt64      │
        ╞═══════╪═════════════╡
        │ 1     ┆ 0           │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2     ┆ 1           │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 3     ┆ 68719476736 │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 4     ┆ 68719476737 │
        ╰───────┴─────────────╯
        <BLANKLINE>
        (Showing first 4 of 4 rows)

    """
    f = native.get_function_from_registry("monotonically_increasing_id")
    return Expression._from_pyexpr(f())


def format(f_string: str, *args: Expression | str) -> Expression:
    """Format a string using the given arguments.

    Args:
        f_string: The format string.
        *args: The arguments to format the string with.

    Returns:
        Expression: A string expression with the formatted result.

    Examples:
        >>> import daft
        >>> from daft.functions import format
        >>> from daft import col
        >>> df = daft.from_pydict({"first_name": ["Alice", "Bob"], "last_name": ["Smith", "Jones"]})
        >>> df = df.with_column("greeting", format("Hello {} {}", col("first_name"), "last_name"))
        >>> df.show()
        ╭────────────┬───────────┬───────────────────╮
        │ first_name ┆ last_name ┆ greeting          │
        │ ---        ┆ ---       ┆ ---               │
        │ Utf8       ┆ Utf8      ┆ Utf8              │
        ╞════════════╪═══════════╪═══════════════════╡
        │ Alice      ┆ Smith     ┆ Hello Alice Smith │
        ├╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ Bob        ┆ Jones     ┆ Hello Bob Jones   │
        ╰────────────┴───────────┴───────────────────╯
        <BLANKLINE>
        (Showing first 2 of 2 rows)
    """
    if f_string.count("{}") != len(args):
        raise ValueError(
            f"Format string {f_string} has {f_string.count('{}')} placeholders but {len(args)} arguments were provided"
        )

    parts = f_string.split("{}")
    exprs = []

    for part, arg in zip(parts, args):
        if part:
            exprs.append(lit(part))

        if isinstance(arg, str):
            exprs.append(col(arg))
        else:
            exprs.append(arg)

    if parts[-1]:
        exprs.append(lit(parts[-1]))

    if not exprs:
        return lit("")

    result = exprs[0]
    for expr in exprs[1:]:
        result = result + expr

    return result



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

logger = logging.getLogger(__name__)

try:
    # When available, subclass from the newer torchdata DataPipes instead of torch Datasets.
    import torchdata

    MAP_DATASET_CLASS = torchdata.datapipes.map.MapDataPipe
    ITER_DATASET_CLASS = torchdata.datapipes.iter.IterDataPipe
except ImportError:
    try:
        import torch

        MAP_DATASET_CLASS = torch.utils.data.Dataset
        ITER_DATASET_CLASS = torch.utils.data.IterableDataset
    except ImportError:
        logger.error("Error when importing Torch. To use PyTorch features, please install torch.")
        raise


class DaftTorchDataset(MAP_DATASET_CLASS):  # type: ignore
    """A wrapper to create a torch map-style Dataset from a Daft pydict of items."""

    def __init__(self, data: dict[str, list[Any]], length: int):
        self.data = data
        self.length = length

    def __len__(self) -> int:
        return self.length

    def __getitem__(self, i: int) -> dict[str, Any]:
        return {key: vallist[i] for (key, vallist) in self.data.items()}


class DaftTorchIterableDataset(ITER_DATASET_CLASS):  # type: ignore
    """A thin wrapper to create a torch IterableDataset from an iterable."""

    def __init__(self, iterable: Iterable[dict[str, Any]]):
        self.iterable = iterable

    def __iter__(self) -> Iterator[dict[str, Any]]:
        return iter(self.iterable)



from __future__ import annotations

import pytest

import daft
from daft.expressions import col, lit
from daft.recordbatch import MicroPartition


@pytest.mark.parametrize("data", [1, "a", True, b"Y", 0.5, None, [1, 2, 3], object()])
def test_broadcast(data):
    table = MicroPartition.from_pydict({"x": [1, 2, 3]})
    new_table = table.eval_expression_list([col("x"), lit(data)])
    assert new_table.to_pydict() == {"x": [1, 2, 3], "literal": [data for _ in range(3)]}


def test_broadcast_fixed_size_list():
    data = [1, 2, 3]
    table = MicroPartition.from_pydict({"x": [1, 2, 3]})
    new_table = table.eval_expression_list(
        [col("x"), lit(data).cast(daft.DataType.fixed_size_list(daft.DataType.int64(), 3))]
    )
    assert new_table.to_pydict() == {"x": [1, 2, 3], "literal": [data for _ in range(3)]}



from __future__ import annotations

from daft.expressions import col
from tests.conftest import assert_df_equals


def test_add_one_to_column(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Creating a new column that is derived from (1 + other_column) and retrieving the top N results."""
    daft_df = daft_df.repartition(repartition_nparts).with_column("unique_key_mod", col("Unique Key") + 1)
    service_requests_csv_pd_df["unique_key_mod"] = service_requests_csv_pd_df["Unique Key"] + 1
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df)


def test_add_one_to_column_name_override(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Creating a new column that is derived from (1 + other_column) and retrieving the top N results."""
    daft_df = (
        daft_df.repartition(repartition_nparts)
        .with_column("Unique Key", col("Unique Key") + 1)
        .with_column("Unique Key", col("Unique Key") + 1)
    )
    service_requests_csv_pd_df["Unique Key"] = service_requests_csv_pd_df["Unique Key"] + 2
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df)


def test_add_one_to_column_limit(daft_df, service_requests_csv_pd_df, with_morsel_size):
    """Creating a new column that is derived from (1 + other_column) and retrieving the top N results."""
    daft_df = daft_df.with_column("unique_key_mod", col("Unique Key") + 1).limit(10)
    service_requests_csv_pd_df["unique_key_mod"] = service_requests_csv_pd_df["Unique Key"] + 1
    service_requests_csv_pd_df = service_requests_csv_pd_df.head(10)
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df)


def test_add_one_twice_to_column(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Creating a new column that is derived from (1 + other_column) and retrieving the top N results."""
    daft_df = daft_df.repartition(repartition_nparts).with_column("unique_key_mod", col("Unique Key") + 1)
    daft_df = daft_df.with_column("unique_key_mod_second", col("unique_key_mod") + 1)
    service_requests_csv_pd_df["unique_key_mod"] = service_requests_csv_pd_df["Unique Key"] + 1
    service_requests_csv_pd_df["unique_key_mod_second"] = service_requests_csv_pd_df["unique_key_mod"] + 1
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df)


def test_difference_cols(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Creating a new column that is derived from 2 other columns and retrieving the top N results."""
    daft_df = daft_df.repartition(repartition_nparts).with_column(
        "unique_key_mod", col("Unique Key") - col("Unique Key")
    )
    service_requests_csv_pd_df["unique_key_mod"] = (
        service_requests_csv_pd_df["Unique Key"] - service_requests_csv_pd_df["Unique Key"]
    )
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df)



from __future__ import annotations

import pytest

import daft
from tests.utils import sort_pydict


def spark_to_daft(spark_df):
    return daft.from_pandas(spark_df.toPandas())


@pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
def test_multicol_joins(join_type, make_spark_df):
    df = make_spark_df(
        {
            "A": [1, 2, 3],
            "B": ["a", "b", "c"],
            "C": [True, False, True],
        }
    )

    joined = df.join(df, on=["A", "B"], how=join_type)
    joined_data = spark_to_daft(joined).to_pydict()
    assert sort_pydict(joined_data, "A", "B", ascending=True) == {
        "A": [1, 2, 3],
        "B": ["a", "b", "c"],
        "C": [True, False, True],
        "right.C": [True, False, True],
    }


def test_columns_after_join(make_spark_df):
    df1 = make_spark_df(
        {
            "A": [1, 2, 3],
        },
    )

    df2 = make_spark_df({"A": [1, 2, 3], "B": [1, 2, 3]})

    joined_df1 = df1.join(df2, df1["A"] == df2["B"])
    joined_df2 = df1.join(df2, df1["A"] == df2["A"])
    joined_df1 = spark_to_daft(joined_df1)
    joined_df2 = spark_to_daft(joined_df2)

    assert set(joined_df1.schema().column_names()) == set(["A", "B", "right.A"])

    assert set(joined_df2.schema().column_names()) == set(["A", "B", "right.A"])


def test_using_columns(make_spark_df):
    df1 = make_spark_df(
        {
            "A": [1, 2, 3],
        },
    )
    df2 = make_spark_df({"A": [1, 2, 3], "B": [1, 2, 3]})

    joined_df = df1.join(df2, "A")
    joined_df = spark_to_daft(joined_df)

    assert set(joined_df.schema().column_names()) == set(["A", "B"])


def test_rename_join_keys_in_dataframe(make_spark_df):
    df1 = make_spark_df({"A": [1, 2], "B": [2, 2]})

    df2 = make_spark_df({"A": [1, 2]})
    joined_df1 = df1.join(df2, (df1["A"] == df2["A"]) & (df1["B"] == df2["A"]))
    joined_df2 = df1.join(df2, (df1["B"] == df2["A"]) & (df1["A"] == df2["A"]))
    joined_df1 = spark_to_daft(joined_df1)
    joined_df2 = spark_to_daft(joined_df2)

    assert set(joined_df1.schema().column_names()) == set(["A", "B", "right.A"])
    assert set(joined_df2.schema().column_names()) == set(["A", "B", "right.A"])


@pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
def test_dupes_join_key(join_type, make_spark_df):
    df = make_spark_df(
        {
            "A": [1, 1, 2, 2, 3, 3],
            "B": ["a", "b", "c", "d", "e", "f"],
        },
    )

    joined = df.join(df, on="A", how=join_type)
    joined = joined.sort(["A", "B", "right.B"])
    joined = spark_to_daft(joined)
    joined_data = joined.to_pydict()

    assert joined_data == {
        "A": [1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3],
        "B": ["a", "a", "b", "b", "c", "c", "d", "d", "e", "e", "f", "f"],
        "right.B": ["a", "b", "a", "b", "c", "d", "c", "d", "e", "f", "e", "f"],
    }


@pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
def test_multicol_dupes_join_key(join_type, make_spark_df):
    df = make_spark_df(
        {
            "A": [1, 1, 2, 2, 3, 3],
            "B": ["a", "a", "b", "b", "c", "d"],
            "C": [1, 0, 1, 0, 1, 0],
        },
    )

    joined = df.join(df, on=["A", "B"], how=join_type)
    joined = joined.sort(["A", "B", "C", "right.C"])
    joined = spark_to_daft(joined)
    joined_data = joined.to_pydict()

    assert joined_data == {
        "A": [1, 1, 1, 1, 2, 2, 2, 2, 3, 3],
        "B": ["a"] * 4 + ["b"] * 4 + ["c", "d"],
        "C": [0, 0, 1, 1, 0, 0, 1, 1, 1, 0],
        "right.C": [0, 1, 0, 1, 0, 1, 0, 1, 1, 0],
    }


@pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
def test_joins_all_same_key(join_type, make_spark_df):
    df = make_spark_df(
        {
            "A": [1] * 4,
            "B": ["a", "b", "c", "d"],
        },
    )

    joined = df.join(df, on="A", how=join_type)
    joined = joined.sort(["A", "B", "right.B"])
    joined = spark_to_daft(joined)
    joined_data = joined.to_pydict()

    assert joined_data == {
        "A": [1] * 16,
        "B": ["a"] * 4 + ["b"] * 4 + ["c"] * 4 + ["d"] * 4,
        "right.B": ["a", "b", "c", "d"] * 4,
    }



from __future__ import annotations

import os


def construct_delta_file_path(scheme: str, table_uri: str, relative_path: str) -> str:
    """Construct full file path for Delta Lake files based on URI scheme.

    Args:
        scheme: The URI scheme (e.g., 's3', 'file', etc.)
        table_uri: The base table URI
        relative_path: The relative path from the Delta log

    Returns:
        The full file path
    """
    if scheme in (
        "s3",
        "s3a",
        "gcs",
        "gs",
        "az",
        "abfs",
        "abfss",
    ):  # object storage does not use os path.join, but instead always uses `/`
        return table_uri.rstrip("/") + "/" + relative_path
    else:
        return os.path.join(table_uri, relative_path)



from __future__ import annotations

import os
from datetime import date, datetime

import pyarrow as pa
import pyarrow.dataset as ds
import pytest

import daft

NUM_ROWS = 100
SCHEMA = pa.schema(
    [
        ("id", pa.int64()),
        ("date_col", pa.date32()),
        ("timestamp_col", pa.timestamp("ns")),
        ("value", pa.int64()),
        ("nullable_int", pa.int64()),
        ("int_col", pa.int64()),
        ("str_col", pa.large_string()),
        ("nullable_str", pa.large_string()),
    ]
)
SAMPLE_DATA = pa.table(
    {
        "id": range(100),
        "str_col": ["str" + str(i % 3) for i in range(100)],
        "int_col": [i % 3 for i in range(100)],
        # We use nonsensical coarse partitions for data and timestamp types to reduce test latency.
        "date_col": [date(2024, 1, i % 3 + 1) for i in range(100)],
        "timestamp_col": [datetime(2024, 1, i % 3 + 1, i % 3) for i in range(100)],
        "value": range(100),
        # Include some nulls.
        "nullable_str": [None if i % 10 == 1 else f"val{i % 3}" for i in range(100)],
        "nullable_int": [None if i % 7 == 1 else i % 3 for i in range(100)],
    },
    schema=SCHEMA,
)


def unify_timestamp(table):
    return table.set_column(
        table.schema.get_field_index("timestamp_col"), "timestamp_col", table["timestamp_col"].cast(pa.timestamp("us"))
    )


def assert_tables_equal(daft_recordbatch, pa_table):
    sorted_pa_table = pa_table.sort_by([("id", "ascending")]).select(SCHEMA.names)
    sorted_pa_table = unify_timestamp(sorted_pa_table)
    sorted_daft_recordbatch = daft_recordbatch.sort_by([("id", "ascending")]).select(SCHEMA.names)
    sorted_daft_recordbatch = unify_timestamp(sorted_daft_recordbatch)
    assert sorted_pa_table == sorted_daft_recordbatch


@pytest.mark.parametrize(
    "partition_by",
    [
        ["str_col"],
        ["int_col"],
        ["date_col"],
        ["timestamp_col"],
        ["nullable_str"],
        ["nullable_int"],
        ["str_col", "int_col"],  # Test multiple partition columns.
        ["nullable_str", "nullable_int"],  # Test multiple partition columns with nulls.
    ],
)
# Pyarrow does not currently support writing partitioned JSON.
@pytest.mark.parametrize("file_format", ["csv", "parquet"])
@pytest.mark.parametrize("filter", [True, False])
def test_hive_pyarrow_daft_compatibility(tmpdir, partition_by, file_format, filter):
    ds.write_dataset(
        SAMPLE_DATA,
        tmpdir,
        format=file_format,
        partitioning=ds.partitioning(pa.schema([SAMPLE_DATA.schema.field(col) for col in partition_by]), flavor="hive"),
    )

    glob_path = os.path.join(tmpdir, "**")
    daft_df = ()
    if file_format == "csv":
        daft_df = daft.read_csv(
            glob_path,
            schema={
                "nullable_str": daft.DataType.string(),
                "nullable_int": daft.DataType.int64(),
            },
            hive_partitioning=True,
        )

    if file_format == "json":
        daft_df = daft.read_json(
            glob_path,
            schema={
                "nullable_str": daft.DataType.string(),
                "nullable_int": daft.DataType.int64(),
            },
            hive_partitioning=True,
        )
    if file_format == "parquet":
        daft_df = daft.read_parquet(
            glob_path,
            schema={
                "nullable_str": daft.DataType.string(),
                "nullable_int": daft.DataType.int64(),
            },
            hive_partitioning=True,
        )
    pa_ds = ds.dataset(
        tmpdir,
        format=file_format,
        partitioning=ds.HivePartitioning.discover(),
        schema=SCHEMA,
    )
    pa_table = pa_ds.to_table()
    if filter:
        first_col = partition_by[0]
        sample_value = SAMPLE_DATA[first_col][0].as_py()
        daft_df = daft_df.where(daft.col(first_col) == sample_value)
        pa_table = pa_ds.to_table(filter=ds.field(first_col) == sample_value)
    assert_tables_equal(daft_df.to_arrow(), pa_table)


@pytest.mark.parametrize(
    "partition_by",
    [
        ["str_col"],
        ["int_col"],
        ["date_col"],
        ["timestamp_col"],
        ["nullable_str"],
        ["nullable_int"],
        ["str_col", "int_col"],  # Test multiple partition columns.
        ["nullable_str", "nullable_int"],  # Test multiple partition columns with nulls.
    ],
)
# TODO(desmond): Daft does not currently have a write_json API.
@pytest.mark.parametrize("file_format", ["csv", "parquet"])
@pytest.mark.parametrize("filter", [True, False])
def test_hive_daft_roundtrip(tmpdir, partition_by, file_format, filter):
    filepath = f"{tmpdir}"
    source = daft.from_arrow(SAMPLE_DATA)

    glob_path = os.path.join(tmpdir, "**")
    target = ()
    if file_format == "csv":
        source.write_csv(filepath, partition_cols=[daft.col(col) for col in partition_by])
        target = daft.read_csv(
            glob_path,
            schema={
                "nullable_str": daft.DataType.string(),
                "nullable_int": daft.DataType.int64(),
            },
            hive_partitioning=True,
        )
        # TODO(desmond): Daft has an inconsistency with handling null string columns when using
        # `from_arrow` vs `read_csv`. For now we read back an unpartitioned CSV table to check the
        # result against.
        plain_filepath = f"{tmpdir}-plain"
        source.write_csv(plain_filepath)
        source = daft.read_csv(
            f"{plain_filepath}/**",
            schema={
                "nullable_str": daft.DataType.string(),
                "nullable_int": daft.DataType.int64(),
            },
        )
    if file_format == "json":
        source.write_json(filepath, partition_cols=[daft.col(col) for col in partition_by])
        target = daft.read_json(
            glob_path,
            schema={
                "nullable_str": daft.DataType.string(),
                "nullable_int": daft.DataType.int64(),
            },
            hive_partitioning=True,
        )
    if file_format == "parquet":
        source.write_parquet(filepath, partition_cols=[daft.col(col) for col in partition_by])
        target = daft.read_parquet(
            glob_path,
            schema={
                "nullable_str": daft.DataType.string(),
                "nullable_int": daft.DataType.int64(),
            },
            hive_partitioning=True,
        )
    if filter:
        first_col = partition_by[0]
        sample_value = SAMPLE_DATA[first_col][0].as_py()
        source = source.where(daft.col(first_col) == sample_value)
        target = target.where(daft.col(first_col) == sample_value)
    assert_tables_equal(target.to_arrow(), source.to_arrow())



from __future__ import annotations

import pytest

import daft


def add_1(df):
    return df.with_column("foo", daft.col("foo") + 1)


def multiply_x(df, x):
    return df.with_column("foo", daft.col("foo") * x)


def concat_dfs(df, df_2):
    return df.concat(df_2)


def invalid_function(df):
    return df.to_pydict()


def test_transform_no_args(make_df):
    df = make_df({"foo": [1, 2, 3]})
    assert df.transform(add_1).to_pydict() == {"foo": [2, 3, 4]}


def test_transform_args(make_df):
    df = make_df({"foo": [1, 2, 3]})
    assert df.transform(multiply_x, 2).to_pydict() == {"foo": [2, 4, 6]}


def test_transform_kwargs(make_df):
    df = make_df({"foo": [1, 2, 3]})
    assert df.transform(multiply_x, x=2).to_pydict() == {"foo": [2, 4, 6]}


def test_transform_multiple_df(make_df):
    df = make_df({"foo": [1, 2, 3]})
    df_2 = make_df({"foo": [4, 5, 6]})
    assert df.transform(concat_dfs, df_2).to_pydict() == {"foo": [1, 2, 3, 4, 5, 6]}


def test_transform_negative_case(make_df):
    df = make_df({"foo": [1, 2, 3]})
    with pytest.raises(AssertionError):
        df.transform(invalid_function)



from __future__ import annotations

import os
import urllib.parse
import uuid
from datetime import date, datetime

import pyarrow as pa
import pytest
from pyarrow import dataset as pads

import daft
from tests.conftest import assert_df_equals
from tests.cookbook.assets import COOKBOOK_DATA_CSV

PYARROW_GE_7_0_0 = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) >= (7, 0, 0)


def test_parquet_write(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)

    pd_df = df.write_parquet(tmp_path)
    read_back_pd_df = daft.read_parquet(tmp_path.as_posix() + "/*.parquet").to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df)

    assert len(pd_df) == 1
    assert pd_df._preview.partition is None
    pd_df.__repr__()
    assert len(pd_df._preview.partition) == 1


def test_parquet_write_with_partitioning(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)

    pd_df = df.write_parquet(tmp_path, partition_cols=["Borough"])

    read_back_pd_df = daft.read_parquet(tmp_path.as_posix() + "/**/*.parquet").to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df)

    assert len(pd_df) == 5
    assert pd_df._preview.partition is None
    pd_df.__repr__()
    assert len(pd_df._preview.partition) == 5


@pytest.mark.parametrize("write_mode", ["append", "overwrite"])
def test_empty_parquet_write_without_partitioning(tmp_path, write_mode, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)
    df = df.where(daft.lit(False))

    # Create a unique path to make sure that the writer is comfortable with nonexistent directories
    path = os.path.join(tmp_path, str(uuid.uuid4()))
    pd_df = df.write_parquet(path, write_mode=write_mode)
    read_back_pd_df = daft.read_parquet(path).to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df)

    assert len(pd_df) == 1
    assert len(pd_df._preview.partition) == 1


@pytest.mark.parametrize("write_mode", ["append", "overwrite"])
def test_empty_parquet_write_with_partitioning(tmp_path, write_mode, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)
    df = df.where(daft.lit(False))

    # Create a unique path to make sure that the writer is comfortable with nonexistent directories
    path = os.path.join(tmp_path, str(uuid.uuid4()))
    output_files = df.write_parquet(path, partition_cols=["Borough"], write_mode=write_mode)
    read_back_pd_df = daft.read_parquet(os.path.join(path, "**/*.parquet")).to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df)

    assert len(output_files) == 1
    assert len(output_files._preview.partition) == 1


def test_parquet_write_with_partitioning_readback_values(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)

    output_files = df.write_parquet(tmp_path, partition_cols=["Borough"])
    output_dict = output_files.to_pydict()
    boroughs = {"QUEENS", "MANHATTAN", "STATEN ISLAND", "BROOKLYN", "BRONX"}
    assert set(output_dict["Borough"]) == boroughs

    for path, bor in zip(output_dict["path"], output_dict["Borough"]):
        assert f"Borough={urllib.parse.quote(bor)}" in path
        read_back = daft.read_parquet(path).to_pydict()
        assert all(b == bor for b in read_back["Borough"])

    read_back_pd_df = daft.read_parquet(tmp_path.as_posix() + "/**/*.parquet").to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df)

    assert len(output_files) == 5
    assert output_files._preview.partition is None
    output_files.__repr__()
    assert len(output_files._preview.partition) == 5


@pytest.mark.parametrize(
    "exp,key,answer",
    [
        (
            daft.col("date").partitioning.days(),
            "date_days",
            [date(2024, 1, 1), date(2024, 2, 1), date(2024, 3, 1), date(2024, 4, 1), date(2024, 5, 1)],
        ),
        (daft.col("date").partitioning.hours(), "date_hours", [473352, 474096, 474792, 475536, 476256]),
        (daft.col("date").partitioning.months(), "date_months", [648, 649, 650, 651, 652]),
        (daft.col("date").partitioning.years(), "date_years", [54]),
    ],
)
def test_parquet_write_with_iceberg_date_partitioning(exp, key, answer, tmp_path, with_morsel_size):
    data = {
        "id": [1, 2, 3, 4, 5],
        "date": [
            datetime(2024, 1, 1),
            datetime(2024, 2, 1),
            datetime(2024, 3, 1),
            datetime(2024, 4, 1),
            datetime(2024, 5, 1),
        ],
    }
    df = daft.from_pydict(data)
    date_files = df.write_parquet(tmp_path, partition_cols=[exp]).sort(by=key)
    output_dict = date_files.to_pydict()
    assert len(output_dict[key]) == len(answer)
    assert output_dict[key] == answer
    read_back_pd_df = daft.read_parquet(tmp_path.as_posix() + "/**/*.parquet").to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df, sort_key="id")


@pytest.mark.parametrize(
    "exp,key,answer",
    [
        (daft.col("id").partitioning.iceberg_bucket(10), "id_bucket", [0, 3, 5, 6, 8]),
        (daft.col("id").partitioning.iceberg_truncate(10), "id_trunc", [0, 10, 20, 40]),
    ],
)
def test_parquet_write_with_iceberg_bucket_and_trunc(exp, key, answer, tmp_path, with_morsel_size):
    data = {
        "id": [1, 12, 23, 24, 45],
        "date": [
            datetime(2024, 1, 1),
            datetime(2024, 2, 1),
            datetime(2024, 3, 1),
            datetime(2024, 4, 1),
            datetime(2024, 5, 1),
        ],
    }
    df = daft.from_pydict(data)
    date_files = df.write_parquet(tmp_path, partition_cols=[exp]).sort(by=key)
    output_dict = date_files.to_pydict()
    assert len(output_dict[key]) == len(answer)
    assert output_dict[key] == answer
    read_back_pd_df = daft.read_parquet(tmp_path.as_posix() + "/**/*.parquet").to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df, sort_key="id")


@pytest.mark.skipif(
    not PYARROW_GE_7_0_0,
    reason="We only use pyarrow datasets 7 for this test",
)
def test_parquet_write_with_null_values(tmp_path, with_morsel_size):
    df = daft.from_pydict({"x": [1, 2, 3, None]})
    df.write_parquet(tmp_path, partition_cols=[df["x"].alias("y")])
    ds = pads.dataset(tmp_path, format="parquet", partitioning=pads.HivePartitioning(pa.schema([("y", pa.int64())])))
    readback = ds.to_table()
    assert readback.to_pydict() == {"x": [1, 2, 3, None], "y": [1, 2, 3, None]}


@pytest.fixture()
def smaller_parquet_target_filesize():
    with daft.execution_config_ctx(parquet_target_filesize=1024):
        yield


@pytest.mark.skipif(
    not PYARROW_GE_7_0_0,
    reason="We only use pyarrow datasets 7 for this test",
)
def test_parquet_write_multifile(tmp_path, smaller_parquet_target_filesize, with_morsel_size):
    data = {"x": list(range(1_000))}
    df = daft.from_pydict(data)
    df2 = df.write_parquet(tmp_path)
    assert len(df2) > 1
    read_back = daft.read_parquet(tmp_path.as_posix() + "/*.parquet").sort(by="x").to_pydict()
    assert read_back == data


@pytest.mark.skipif(
    not PYARROW_GE_7_0_0,
    reason="We only use pyarrow datasets 7 for this test",
)
def test_parquet_write_multifile_with_partitioning(tmp_path, smaller_parquet_target_filesize, with_morsel_size):
    data = {"x": list(range(1_000))}
    df = daft.from_pydict(data)
    df2 = df.write_parquet(tmp_path, partition_cols=[df["x"].alias("y") % 2])
    assert len(df2) >= 4
    ds = pads.dataset(tmp_path, format="parquet", partitioning=pads.HivePartitioning(pa.schema([("y", pa.int64())])))
    readback = ds.to_table()
    readback = readback.sort_by("x").to_pydict()
    assert readback["x"] == data["x"]
    assert readback["y"] == [y % 2 for y in data["x"]]


def test_parquet_write_with_some_empty_partitions(tmp_path, with_morsel_size):
    data = {"x": [1, 2, 3], "y": ["a", "b", "c"]}
    daft.from_pydict(data).into_partitions(4).write_parquet(tmp_path)

    read_back = daft.read_parquet(tmp_path.as_posix() + "/**/*.parquet").sort("x").to_pydict()
    assert read_back == data


def test_parquet_partitioned_write_with_some_empty_partitions(tmp_path, with_morsel_size):
    data = {"x": [1, 2, 3], "y": ["a", "b", "c"]}
    output_files = daft.from_pydict(data).into_partitions(4).write_parquet(tmp_path, partition_cols=["x"])

    assert len(output_files) == 3

    read_back = daft.read_parquet(tmp_path.as_posix() + "/**/*.parquet").sort("x").to_pydict()
    assert read_back == data


def test_csv_write(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)

    pd_df = df.write_csv(tmp_path)

    read_back_pd_df = daft.read_csv(tmp_path.as_posix() + "/*.csv").to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df)

    assert len(pd_df) == 1
    assert pd_df._preview.partition is None
    pd_df.__repr__()
    assert len(pd_df._preview.partition) == 1


def test_csv_write_with_partitioning(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)
    schema = df.schema()
    names = schema.column_names()
    types = {}
    for n in names:
        types[n] = schema[n].dtype

    pd_df = df.write_csv(tmp_path, partition_cols=["Borough"]).to_pandas()
    read_back_pd_df = daft.read_csv(tmp_path.as_posix() + "/**/*.csv", schema=types).to_pandas()
    assert_df_equals(df.to_pandas().fillna(""), read_back_pd_df.fillna(""))

    assert len(pd_df) == 5


def test_empty_csv_write(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)
    df = df.where(daft.lit(False))

    schema = df.schema()
    names = schema.column_names()
    types = {}
    for n in names:
        types[n] = schema[n].dtype

    pd_df = df.write_csv(tmp_path)
    read_back_pd_df = daft.read_csv(tmp_path.as_posix() + "/*.csv", schema=types).to_pandas()
    assert_df_equals(df.to_pandas().fillna(""), read_back_pd_df.fillna(""))

    assert len(pd_df) == 1
    assert len(pd_df._preview.partition) == 1


def test_empty_csv_write_with_partitioning(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)
    df = df.where(daft.lit(False))

    schema = df.schema()
    names = schema.column_names()
    types = {}
    for n in names:
        types[n] = schema[n].dtype

    pd_df = df.write_csv(tmp_path, partition_cols=["Borough"])
    read_back_pd_df = daft.read_csv(tmp_path.as_posix() + "/**/*.csv", schema=types).to_pandas()
    assert_df_equals(df.to_pandas().fillna(""), read_back_pd_df.fillna(""))

    assert len(pd_df) == 1
    assert len(pd_df._preview.partition) == 1


def test_csv_write_with_some_empty_partitions(tmp_path, with_morsel_size):
    data = {"x": [1, 2, 3], "y": ["a", "b", "c"]}
    daft.from_pydict(data).into_partitions(4).write_csv(tmp_path)

    read_back = daft.read_csv(tmp_path.as_posix() + "/**/*.csv").sort("x").to_pydict()
    assert read_back == data


def test_csv_partitioned_write_with_some_empty_partitions(tmp_path, with_morsel_size):
    data = {"x": [1, 2, 3], "y": ["a", "b", "c"]}
    output_files = daft.from_pydict(data).into_partitions(4).write_csv(tmp_path, partition_cols=["x"])

    assert len(output_files) == 3

    read_back = daft.read_csv(tmp_path.as_posix() + "/**/*.csv").sort("x").to_pydict()
    assert read_back == data



from __future__ import annotations

import pytest


@pytest.mark.integration()
def test_load_volume(local_unity_catalog):
    json_files = local_unity_catalog.load_volume("unity.default.json_files")
    assert json_files.volume_info.full_name == "unity.default.json_files"

    txt_files = local_unity_catalog.load_volume("unity.default.txt_files")
    assert txt_files.volume_info.full_name == "unity.default.txt_files"

    with pytest.raises(Exception):
        local_unity_catalog.load_volume("unity.default.nonexistent_volume")


@pytest.mark.integration()
@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
def test_read_volume(make_df, local_unity_catalog, repartition_nparts, with_morsel_size):
    df = make_df(
        {
            "id": [1, 2, 3, 4, 5],
            "files": [
                "vol+dbfs:/Volumes/unity/default/txt_files/a.txt",
                "vol+dbfs:/Volumes/unity/default/json_files/c.json",
                "vol+dbfs:/Volumes/unity/default/json_files/d.json",
                "dbfs:/Volumes/unity/default/json_files/dir1/e.json",
                "dbfs:/Volumes/unity/default/txt_files/b.txt",
            ],
        },
        repartition=repartition_nparts,
    )

    io_config = local_unity_catalog.to_io_config()

    df = df.select("id", df["files"].url.download(io_config=io_config))

    result = df.sort("id").to_pydict()
    assert result == {
        "id": [1, 2, 3, 4, 5],
        "files": [
            b"Managed volume file name a.txt",
            b'{\n  "marks" :[\n    {"name" :  "a" , "score" :  20},\n    {"name" :  "b" , "score" :  30},\n    {"name" :  "c" , "score" :  40},\n    {"name" :  "d" , "score" :  50}\n  ]\n}\n',
            b'{\n  "type": "object",\n  "properties": {\n    "name": {\n      "type": "string"\n    },\n    "age": {\n      "type": "integer"\n    },\n    "address": {\n      "type": "string"\n    }\n  }\n}\n',
            b"[1,2,3,4]\n",
            b"Managed volume file name b.txt",
        ],
    }



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

from daft.context import get_context
from daft.datatype import DataType
from daft.io.common import _get_schema_from_dict
from daft.recordbatch.micropartition import MicroPartition
from daft.recordbatch.partitioning import PartitionedTable, partition_strings_to_path

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping

    from deltalake.writer import AddAction

    from daft.daft import IOConfig
    from daft.dependencies import pa, pads, pafs

try:
    import deltalake
    from packaging.version import parse

    if parse(deltalake.__version__) < parse("1.0.0"):
        from deltalake.writer import AddAction
    else:
        from deltalake.transaction import AddAction
except ImportError:
    pass


def sanitize_table_for_deltalake(
    table: MicroPartition, large_dtypes: bool, partition_keys: list[str] | None = None
) -> pa.Table:
    arrow_table = table.to_arrow()

    # Remove partition keys from the table since they are already encoded as keys
    if partition_keys is not None:
        arrow_table = arrow_table.drop_columns(partition_keys)

    arrow_batch = convert_pa_schema_to_delta(arrow_table.schema, large_dtypes)
    return arrow_table.cast(arrow_batch)


def partitioned_table_to_deltalake_iter(
    partitioned: PartitionedTable, large_dtypes: bool
) -> Iterator[tuple[pa.Table, str, dict[str, str | None]]]:
    """Iterates over partitions, yielding each partition as an Arrow table, along with their respective paths and partition values."""
    partition_values = partitioned.partition_values()

    if partition_values:
        partition_keys = partition_values.column_names()
        partition_strings = partitioned.partition_values_str()
        assert partition_strings is not None

        for part_table, part_strs in zip(partitioned.partitions(), partition_strings.to_pylist()):
            part_path = partition_strings_to_path("", part_strs)
            converted_arrow_table = sanitize_table_for_deltalake(part_table, large_dtypes, partition_keys)
            yield converted_arrow_table, part_path, part_strs
    else:
        converted_arrow_table = sanitize_table_for_deltalake(partitioned.table, large_dtypes)
        yield converted_arrow_table, "/", {}


def get_file_stats_from_metadata(
    metadata: Any,
) -> dict[str, int | dict[str, Any]]:
    from math import inf

    stats = {
        "numRecords": metadata.num_rows,
        "minValues": {},
        "maxValues": {},
        "nullCount": {},
    }

    def iter_groups(metadata: Any) -> Iterator[Any]:
        for i in range(metadata.num_row_groups):
            if metadata.row_group(i).num_rows > 0:
                yield metadata.row_group(i)

    for column_idx in range(metadata.num_columns):
        name = metadata.row_group(0).column(column_idx).path_in_schema

        # If stats missing, then we can't know aggregate stats
        if all(group.column(column_idx).is_stats_set for group in iter_groups(metadata)):
            stats["nullCount"][name] = sum(
                group.column(column_idx).statistics.null_count for group in iter_groups(metadata)
            )

            # Min / max may not exist for some column types, or if all values are null
            if any(group.column(column_idx).statistics.has_min_max for group in iter_groups(metadata)):
                # Min and Max are recorded in physical type, not logical type
                # https://stackoverflow.com/questions/66753485/decoding-parquet-min-max-statistics-for-decimal-type
                # TODO: Add logic to decode physical type for DATE, DECIMAL

                minimums = (group.column(column_idx).statistics.min for group in iter_groups(metadata))
                # If some row groups have all null values, their min and max will be null too.
                min_value = min(minimum for minimum in minimums if minimum is not None)
                # Infinity cannot be serialized to JSON, so we skip it. Saying
                # min/max is infinity is equivalent to saying it is null, anyways.
                if min_value != -inf:
                    stats["minValues"][name] = min_value
                maximums = (group.column(column_idx).statistics.max for group in iter_groups(metadata))
                max_value = max(maximum for maximum in maximums if maximum is not None)
                if max_value != inf:
                    stats["maxValues"][name] = max_value
    return stats


class DeltaJSONEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        from datetime import date, datetime
        from decimal import Decimal

        if isinstance(obj, bytes):
            return obj.decode("unicode_escape", "backslashreplace")
        elif isinstance(obj, date):
            return obj.isoformat()
        elif isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, Decimal):
            return str(obj)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def make_deltalake_add_action(
    path: str,
    metadata: Any,
    size: int,
    partition_values: Mapping[str, str | None],
) -> AddAction:
    import json
    from datetime import datetime

    stats = get_file_stats_from_metadata(metadata)

    # remove leading slash
    path = path[1:] if path.startswith("/") else path
    return AddAction(
        path,
        size,
        partition_values,
        int(datetime.now().timestamp() * 1000),
        True,
        json.dumps(stats, cls=DeltaJSONEncoder),
    )


def make_deltalake_fs(path: str, io_config: IOConfig | None = None) -> pafs.PyFileSystem:
    from deltalake.fs import DeltaStorageHandler
    from pyarrow.fs import PyFileSystem

    from daft.io.object_store_options import io_config_to_storage_options

    io_config = get_context().daft_planning_config.default_io_config if io_config is None else io_config
    storage_options = io_config_to_storage_options(io_config, path)
    return PyFileSystem(DeltaStorageHandler(path, storage_options))


class DeltaLakeWriteVisitors:
    class FileVisitor:
        def __init__(
            self,
            parent: DeltaLakeWriteVisitors,
            partition_values: dict[str, str | None],
        ):
            self.parent = parent
            self.partition_values = partition_values

        def __call__(self, written_file: pads.WrittenFile) -> None:
            from daft.utils import get_arrow_version

            # PyArrow added support for size in 9.0.0
            if get_arrow_version() >= (9, 0, 0):
                size = written_file.size
            elif self.parent.fs is not None:
                size = self.parent.fs.get_file_info([written_file.path])[0].size
            else:
                size = 0

            add_action = make_deltalake_add_action(
                written_file.path, written_file.metadata, size, self.partition_values
            )

            self.parent.add_actions.append(add_action)

    def __init__(self, fs: pafs.FileSystem):
        self.add_actions: list[AddAction] = []
        self.fs = fs

    def visitor(self, partition_values: dict[str, str | None]) -> DeltaLakeWriteVisitors.FileVisitor:
        return self.FileVisitor(self, partition_values)

    def to_metadata(self) -> MicroPartition:
        col_name = "add_action"
        if len(self.add_actions) == 0:
            return MicroPartition.empty(_get_schema_from_dict({col_name: DataType.python()}))
        return MicroPartition.from_pydict({col_name: self.add_actions})


def convert_pa_schema_to_delta(schema: pa.Schema, large_dtypes: bool) -> pa.Schema:
    import deltalake
    from packaging.version import parse

    deltalake_version = parse(deltalake.__version__)

    if deltalake_version < parse("0.19.0"):
        from deltalake.schema import _convert_pa_schema_to_delta

        return _convert_pa_schema_to_delta(schema, large_dtypes=large_dtypes)
    elif deltalake_version < parse("1.0.0"):
        from deltalake.schema import ArrowSchemaConversionMode, _convert_pa_schema_to_delta

        schema_conversion_mode = ArrowSchemaConversionMode.LARGE if large_dtypes else ArrowSchemaConversionMode.NORMAL
        return _convert_pa_schema_to_delta(schema, schema_conversion_mode=schema_conversion_mode)
    else:
        # deltalake>=1.0.0 passes through Arrow data types without modification
        # https://delta-io.github.io/delta-rs/upgrade-guides/guide-1.0.0/#large_dtypes-removed
        return schema


def create_table_with_add_actions(
    table_uri: str,
    schema: pa.Schema,
    add_actions: list[AddAction],
    mode: str,
    partition_by: list[str],
    name: str | None,
    description: str | None,
    configuration: Mapping[str, str | None] | None,
    storage_options: dict[str, str] | None,
    custom_metadata: dict[str, str] | None,
) -> None:
    import deltalake
    from packaging.version import parse

    if parse(deltalake.__version__) < parse("1.0.0"):
        from deltalake.writer import write_deltalake_pyarrow

        return write_deltalake_pyarrow(
            table_uri,
            schema,
            add_actions,
            mode,
            partition_by,
            name,
            description,
            configuration,
            storage_options,
            custom_metadata,
        )
    else:
        from deltalake import CommitProperties
        from deltalake.transaction import create_table_with_add_actions

        return create_table_with_add_actions(
            table_uri,
            deltalake.Schema.from_arrow(schema),
            add_actions,
            mode,
            partition_by,
            name,
            description,
            configuration,
            storage_options,
            CommitProperties(custom_metadata=custom_metadata),
        )


__all__ = ["AddAction"]



from __future__ import annotations

import logging
import os
import signal
import subprocess
from collections.abc import Generator
from contextlib import _GeneratorContextManager, contextmanager
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Generator

    from viztracer import VizTracer

ACTIVE = False


logger = logging.getLogger(__name__)


@contextmanager
def profiler(filename: str) -> Generator[VizTracer | None, Any, None]:
    if int(os.environ.get("DAFT_PROFILING", 0)) == 1:
        from viztracer import VizTracer, get_tracer

        global ACTIVE
        if not ACTIVE:
            tracer = get_tracer()
            if tracer is None or tracer.output_file != filename:
                tracer = VizTracer(output_file=filename, tracer_entries=10_000_000)
            ACTIVE = True
            with tracer:
                yield tracer
            ACTIVE = False
            return
        else:
            tracer = get_tracer()
            logger.warning(
                "profiler(%s) not created. Another profiler(%s) is already active.",
                filename,
                tracer.output_file,
            )

    yield None
    return


import time


@contextmanager
def timingcontext(name: str) -> Generator[None, Any, None]:
    from viztracer import get_tracer

    tracer = get_tracer()

    try:
        logger.debug("log_event:enter:%s", name)
        start = time.time()
        if tracer is not None:
            with tracer.log_event(name) as event:
                yield event
        else:
            yield None
    finally:
        end = time.time()

        logger.debug("log_event:%s:%sms", name, f"{(end-start)*1000:.3f}")


def log_event(name: str) -> _GeneratorContextManager[None]:
    return timingcontext(name)


ENABLE_SAMPLY_PROFILING = os.getenv("DAFT_DEV_ENABLE_SAMPLY_PROFILING", "false").lower() == "true"


@contextmanager
def profile() -> Generator[None, None, None]:
    """Context manager to profile the current process with samply."""
    if not ENABLE_SAMPLY_PROFILING:
        yield
        return

    current_pid = os.getpid()

    # Run the profiler from a bash script to avoid issues with the profiler process being unable to receive signals.
    # https://github.com/ray-project/ray/issues/31805
    cmd = [
        "bash",
        "-c",
        f"""
        samply record --pid {current_pid} -s -o /tmp/ray/session_latest/logs/daft/samply_{current_pid}_{time.time()}.json.gz
        """,
    ]

    try:
        logger.debug("Starting samply profiler...")
        # Start the profiler process
        profiler_process = subprocess.Popen(cmd)

        # Yield control back to the context
        yield

    finally:
        # Stop profiling by sending SIGINT (Ctrl+C) to samply
        if profiler_process.poll() is None:  # Process is still running
            logger.debug("Stopping samply profiler...")
            profiler_process.send_signal(signal.SIGINT)
            logger.debug("Waiting for samply profiler to finish...")
            profiler_process.wait()
            logger.debug("Samply profiler finished")



from __future__ import annotations

import pytest

import daft
from daft import DataFrame


def test_dataframe_getitem_single(valid_data: list[dict[str, float]]) -> None:
    df = daft.from_pylist(valid_data)
    expanded_df = df.with_column("foo", df["sepal_length"] + df["sepal_width"])
    # TODO(jay): Test that the expression with name "foo" is equal to the expected expression, except for the IDs of the columns

    assert expanded_df.column_names == df.column_names + ["foo"]
    assert df.select(df["sepal_length"]).column_names == ["sepal_length"]


def test_dataframe_getitem_single_bad(valid_data: list[dict[str, float]]) -> None:
    df = daft.from_pylist(valid_data)
    with pytest.raises(ValueError, match="foo does not exist in schema"):
        df["foo"]

    with pytest.raises(ValueError, match="bounds"):
        df[-100]

    with pytest.raises(ValueError, match="bounds"):
        df[100]


def test_dataframe_getitem_multiple_bad(valid_data: list[dict[str, float]]) -> None:
    df = daft.from_pylist(valid_data)
    with pytest.raises(ValueError, match="not found"):
        df["foo", "bar"]

    with pytest.raises(ValueError, match="bounds"):
        df[-100, -200]

    with pytest.raises(ValueError, match="bounds"):
        df[100, 200]

    with pytest.raises(ValueError, match="indexing type"):
        df[[{"a": 1}]]

    class A: ...

    with pytest.raises(ValueError, match="indexing type"):
        df[A()]


def test_dataframe_getitem_multiple(valid_data: list[dict[str, float]]) -> None:
    df = daft.from_pylist(valid_data)
    expanded_df = df.with_column("foo", sum(df["sepal_length", "sepal_width"].columns))
    # TODO(jay): Test that the expression with name "foo" is equal to the expected expression, except for the IDs of the columns
    assert expanded_df.column_names == df.column_names + ["foo"]
    assert isinstance(df["sepal_length", "sepal_width"], DataFrame)
    assert df["sepal_length", "sepal_width"].column_names == ["sepal_length", "sepal_width"]


def test_dataframe_getitem_slice(valid_data: list[dict[str, float]]) -> None:
    df = daft.from_pylist(valid_data)
    slice_df = df[:]
    assert df.column_names == slice_df.column_names


def test_dataframe_getitem_slice_rev(valid_data: list[dict[str, float]]) -> None:
    df = daft.from_pylist(valid_data)
    slice_df = df[::-1]
    assert df.column_names == slice_df.column_names[::-1]



from __future__ import annotations

from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_utf8_like():
    table = MicroPartition.from_pydict({"col": ["foo", None, "barBaz", "quux", "1"]})
    result = table.eval_expression_list([col("col").str.like("foo%")])
    assert result.to_pydict() == {"col": [True, None, False, False, False]}



from __future__ import annotations

import datetime
import random

import pandas as pd
import pytest

import daft
from daft import Window, col
from daft.expressions import interval
from daft.functions import dense_rank, rank, row_number
from daft.sql.sql import SQLCatalog
from tests.conftest import assert_df_equals, get_tests_daft_runner_name

pytestmark = pytest.mark.skipif(
    get_tests_daft_runner_name() == "ray",
    # and get_context().daft_execution_config.use_experimental_distributed_engine is False,
    reason="requires Native Runner to be in use",
)


def test_row_number_window_function():
    """Test SQL ROW_NUMBER() window function over partitions with ordering."""
    random.seed(42)

    data = {"category": [], "value": []}

    for category in ["A", "B", "C"]:
        values = random.sample(range(1, 1001), 100)

        for value in values:
            data["category"].append(category)
            data["value"].append(value)

    df = daft.from_pydict(data)

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            value,
            ROW_NUMBER() OVER(
                PARTITION BY category
                ORDER BY value
            ) AS row_num
        FROM test_data
        ORDER BY category, value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("value")

    daft_result = df.with_column("row_num", row_number().over(window_spec)).sort(["category", "value"]).collect()

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "value"])


def test_rank_window_function():
    """Test SQL RANK() window function over partitions with ordering."""
    df = daft.from_pydict(
        {
            "category": ["A", "A", "A", "B", "B", "B", "C", "C", "C"],
            "value": [1, 1, 2, 2, 2, 3, 3, 4, 4],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            value,
            RANK() OVER(
                PARTITION BY category
                ORDER BY value
            ) AS rank_val
        FROM test_data
        ORDER BY category, value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("value")

    daft_result = df.with_column("rank_val", rank().over(window_spec)).sort(["category", "value"]).collect()

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "value"])


def test_dense_rank_window_function():
    """Test SQL DENSE_RANK() window function over partitions with ordering."""
    df = daft.from_pydict(
        {
            "category": ["A", "A", "A", "B", "B", "B", "C", "C", "C"],
            "value": [1, 1, 2, 2, 2, 3, 3, 4, 4],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            value,
            DENSE_RANK() OVER(
                PARTITION BY category
                ORDER BY value
            ) AS dense_rank_val
        FROM test_data
        ORDER BY category, value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("value")

    daft_result = df.with_column("dense_rank_val", dense_rank().over(window_spec)).sort(["category", "value"]).collect()

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "value"])


def test_lag_with_one_arg():
    """Test SQL LAG() with one argument."""
    df = daft.from_pydict(
        {
            "category": ["A", "A", "A", "B", "B", "B"],
            "value": [1, 2, 3, 4, 5, 6],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            value,
            LAG(value) OVER(
                PARTITION BY category
                ORDER BY value
            ) AS lagged
        FROM test_data
        ORDER BY category, value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("value")

    daft_result = df.with_column("lagged", col("value").lag(1).over(window_spec)).sort(["category", "value"]).collect()

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "value"])


def test_lag_with_two_args():
    """Test SQL LAG() with two arguments (value, offset)."""
    df = daft.from_pydict(
        {
            "category": ["A", "A", "A", "B", "B", "B"],
            "value": [1, 2, 3, 4, 5, 6],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            value,
            LAG(value, 2) OVER(
                PARTITION BY category
                ORDER BY value
            ) AS lagged
        FROM test_data
        ORDER BY category, value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("value")

    daft_result = df.with_column("lagged", col("value").lag(2).over(window_spec)).sort(["category", "value"]).collect()

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "value"])


def test_lag_with_three_args():
    """Test SQL LAG() with three arguments (value, offset, default)."""
    df = daft.from_pydict(
        {
            "category": ["A", "A", "A", "B", "B", "B"],
            "value": [1, 2, 3, 4, 5, 6],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            value,
            LAG(value, 2, 999) OVER(
                PARTITION BY category
                ORDER BY value
            ) AS lagged
        FROM test_data
        ORDER BY category, value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("value")

    daft_result = (
        df.with_column("lagged", col("value").lag(2, 999).over(window_spec)).sort(["category", "value"]).collect()
    )

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "value"])


def test_lead_with_one_arg():
    """Test SQL LEAD() with one argument."""
    df = daft.from_pydict(
        {
            "category": ["A", "A", "A", "B", "B", "B"],
            "value": [1, 2, 3, 4, 5, 6],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            value,
            LEAD(value) OVER(
                PARTITION BY category
                ORDER BY value
            ) AS lead_val
        FROM test_data
        ORDER BY category, value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("value")

    daft_result = (
        df.with_column("lead_val", col("value").lead(1).over(window_spec)).sort(["category", "value"]).collect()
    )

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "value"])


def test_lead_with_two_args():
    """Test SQL LEAD() with two arguments (value, offset)."""
    df = daft.from_pydict(
        {
            "category": ["A", "A", "A", "B", "B", "B"],
            "value": [1, 2, 3, 4, 5, 6],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            value,
            LEAD(value, 2) OVER(
                PARTITION BY category
                ORDER BY value
            ) AS lead_val
        FROM test_data
        ORDER BY category, value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("value")

    daft_result = (
        df.with_column("lead_val", col("value").lead(2).over(window_spec)).sort(["category", "value"]).collect()
    )

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "value"])


def test_lead_with_three_args():
    """Test SQL LEAD() with three arguments (value, offset, default)."""
    df = daft.from_pydict(
        {
            "category": ["A", "A", "A", "B", "B", "B"],
            "value": [1, 2, 3, 4, 5, 6],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            value,
            LEAD(value, 2, 999) OVER(
                PARTITION BY category
                ORDER BY value
            ) AS lead_val
        FROM test_data
        ORDER BY category, value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("value")

    daft_result = (
        df.with_column("lead_val", col("value").lead(2, 999).over(window_spec)).sort(["category", "value"]).collect()
    )

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "value"])


def test_multiple_window_functions():
    """Test using multiple window functions together."""
    df = daft.from_pydict(
        {
            "category": ["A", "A", "A", "B", "B", "B"],
            "value": [1, 2, 3, 4, 5, 6],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            value,
            ROW_NUMBER() OVER(PARTITION BY category ORDER BY value) AS row_num,
            RANK() OVER(PARTITION BY category ORDER BY value) AS rank_val,
            DENSE_RANK() OVER(PARTITION BY category ORDER BY value) AS dense_rank_val,
            LAG(value) OVER(PARTITION BY category ORDER BY value) AS lag_val,
            LEAD(value) OVER(PARTITION BY category ORDER BY value) AS lead_val
        FROM test_data
        ORDER BY category, value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("value")

    daft_result = (
        df.with_column("row_num", row_number().over(window_spec))
        .with_column("rank_val", rank().over(window_spec))
        .with_column("dense_rank_val", dense_rank().over(window_spec))
        .with_column("lag_val", col("value").lag(1).over(window_spec))
        .with_column("lead_val", col("value").lead(1).over(window_spec))
        .sort(["category", "value"])
        .collect()
    )

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "value"])


def test_row_number_and_running_sum_window_functions():
    """Test SQL ROW_NUMBER() and SUM() window functions over partitions with ordering."""
    random.seed(42)

    data = {"category": [], "value": []}

    for category in ["A", "B", "C"]:
        values = random.sample(range(1, 1001), 100)

        for value in values:
            data["category"].append(category)
            data["value"].append(value)

    df = daft.from_pydict(data)

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            value,
            ROW_NUMBER() OVER(
                PARTITION BY category
                ORDER BY value
            ) AS row_num,
            SUM(value) OVER(
                PARTITION BY category
                ORDER BY value
                ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW
            ) AS running_sum
        FROM test_data
        ORDER BY category, value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("value")
    running_window = (
        Window().partition_by("category").order_by("value").rows_between(Window.unbounded_preceding, Window.current_row)
    )

    daft_result = (
        df.with_column("row_num", row_number().over(window_spec))
        .with_column("running_sum", col("value").sum().over(running_window))
        .sort(["category", "value"])
        .collect()
    )

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "value"])


def test_range_window_sql():
    """Test SQL window functions with range frame type."""
    random.seed(50)

    data = {"category": [], "ts": [], "value": []}
    possible_timestamps = random.sample(range(1000), 300)
    possible_timestamps.sort()

    for category in ["A", "B"]:
        timestamps = possible_timestamps.copy()
        values = [random.randint(1, 100) for _ in range(len(timestamps))]

        for ts, value in zip(timestamps, values):
            data["category"].append(category)
            data["ts"].append(ts)
            data["value"].append(value)

    df = daft.from_pydict(data)

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            ts,
            value,
            SUM(value) OVER(
                PARTITION BY category
                ORDER BY ts
                RANGE BETWEEN 2 PRECEDING AND 2 FOLLOWING
            ) AS range_sum,
            AVG(value) OVER(
                PARTITION BY category
                ORDER BY ts
                RANGE BETWEEN 2 PRECEDING AND 2 FOLLOWING
            ) AS range_avg,
            MIN(value) OVER(
                PARTITION BY category
                ORDER BY ts
                RANGE BETWEEN 2 PRECEDING AND 2 FOLLOWING
            ) AS range_min,
            MAX(value) OVER(
                PARTITION BY category
                ORDER BY ts
                RANGE BETWEEN 2 PRECEDING AND 2 FOLLOWING
            ) AS range_max
        FROM test_data
        ORDER BY category, ts
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("ts", desc=False).range_between(-2, 2)

    daft_result = (
        df.with_column("range_sum", col("value").sum().over(window_spec))
        .with_column("range_avg", col("value").mean().over(window_spec))
        .with_column("range_min", col("value").min().over(window_spec))
        .with_column("range_max", col("value").max().over(window_spec))
        .sort(["category", "ts"])
        .collect()
    )

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "ts"], check_dtype=False)


def test_range_window_desc_sql():
    """Test SQL window functions with range frame type and descending order."""
    random.seed(53)

    data = {"category": [], "ts": [], "value": []}
    possible_timestamps = random.sample(range(1000), 300)
    possible_timestamps.sort()

    for category in ["A", "B"]:
        timestamps = possible_timestamps.copy()
        values = [random.randint(1, 100) for _ in range(len(timestamps))]

        for ts, value in zip(timestamps, values):
            data["category"].append(category)
            data["ts"].append(ts)
            data["value"].append(value)

    df = daft.from_pydict(data)

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            ts,
            value,
            SUM(value) OVER(
                PARTITION BY category
                ORDER BY ts DESC
                RANGE BETWEEN 2 PRECEDING AND 2 FOLLOWING
            ) AS range_sum,
            AVG(value) OVER(
                PARTITION BY category
                ORDER BY ts DESC
                RANGE BETWEEN 2 PRECEDING AND 2 FOLLOWING
            ) AS range_avg,
            MIN(value) OVER(
                PARTITION BY category
                ORDER BY ts DESC
                RANGE BETWEEN 2 PRECEDING AND 2 FOLLOWING
            ) AS range_min,
            MAX(value) OVER(
                PARTITION BY category
                ORDER BY ts DESC
                RANGE BETWEEN 2 PRECEDING AND 2 FOLLOWING
            ) AS range_max
        FROM test_data
        ORDER BY category, ts
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("ts", desc=True).range_between(-2, 2)

    daft_result = (
        df.with_column("range_sum", col("value").sum().over(window_spec))
        .with_column("range_avg", col("value").mean().over(window_spec))
        .with_column("range_min", col("value").min().over(window_spec))
        .with_column("range_max", col("value").max().over(window_spec))
        .sort(["category", "ts"])
        .collect()
    )

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "ts"], check_dtype=False)


def test_range_window_with_dates():
    """Test SQL window functions with range frame type and date/timestamp data."""
    random.seed(54)

    base_date = datetime.datetime(2023, 1, 1)
    data = {"category": [], "date": [], "value": []}

    for category in ["A", "B"]:
        dates = sorted([random.randint(0, 1000) for _ in range(700)])
        values = random.sample(range(1, 1000), 700)
        dates = [base_date + datetime.timedelta(days=date) for date in dates]

        for date, value in zip(dates, values):
            data["category"].append(category)
            data["date"].append(date)
            data["value"].append(value)

    pdf = pd.DataFrame(data)
    df = daft.from_pandas(pdf)

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            category,
            date,
            value,
            SUM(value) OVER(
                PARTITION BY category
                ORDER BY date
                RANGE BETWEEN INTERVAL '3' DAY PRECEDING AND INTERVAL '3' DAY FOLLOWING
            ) AS date_range_sum,
            AVG(value) OVER(
                PARTITION BY category
                ORDER BY date
                RANGE BETWEEN INTERVAL '3' DAY PRECEDING AND INTERVAL '3' DAY FOLLOWING
            ) AS date_range_avg
        FROM test_data
        ORDER BY category, date
        """,
        catalog=catalog,
    ).collect()

    window_spec = Window().partition_by("category").order_by("date").range_between(interval(days=-3), interval(days=3))

    daft_result = (
        df.with_column("date_range_sum", col("value").sum().over(window_spec))
        .with_column("date_range_avg", col("value").mean().over(window_spec))
        .sort(["category", "date"])
        .collect()
    )

    assert_df_equals(
        sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["category", "date", "value"], check_dtype=False
    )


def test_order_by_only_row_number():
    """Test SQL ROW_NUMBER() window function with only ORDER BY (no PARTITION BY)."""
    df = daft.from_pydict(
        {
            "value": [100, 200, 50, 500, 125, 300, 250, 150],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            value,
            ROW_NUMBER() OVER(
                ORDER BY value
            ) AS row_num
        FROM test_data
        ORDER BY value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().order_by("value")

    daft_result = df.with_column("row_num", row_number().over(window_spec)).sort(["value"]).collect()

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["value"])


def test_order_by_only_rank():
    """Test SQL RANK() window function with only ORDER BY (no PARTITION BY)."""
    df = daft.from_pydict(
        {
            "value": [1, 1, 2, 2, 2, 3, 3, 4, 4],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            value,
            RANK() OVER(
                ORDER BY value
            ) AS rank_val
        FROM test_data
        ORDER BY value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().order_by("value")

    daft_result = df.with_column("rank_val", rank().over(window_spec)).sort(["value"]).collect()

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["value"])


def test_order_by_only_dense_rank():
    """Test SQL DENSE_RANK() window function with only ORDER BY (no PARTITION BY)."""
    df = daft.from_pydict(
        {
            "value": [1, 1, 2, 2, 2, 3, 3, 4, 4],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            value,
            DENSE_RANK() OVER(
                ORDER BY value
            ) AS dense_rank_val
        FROM test_data
        ORDER BY value
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().order_by("value")

    daft_result = df.with_column("dense_rank_val", dense_rank().over(window_spec)).sort(["value"]).collect()

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["value"])


def test_order_by_only_desc():
    """Test window functions with ORDER BY DESC."""
    df = daft.from_pydict(
        {
            "value": [1, 2, 3, 4, 5],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            value,
            ROW_NUMBER() OVER(ORDER BY value DESC) AS row_num,
            RANK() OVER(ORDER BY value DESC) AS rank_val,
            DENSE_RANK() OVER(ORDER BY value DESC) AS dense_rank_val
        FROM test_data
        ORDER BY value DESC
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().order_by("value", desc=True)

    daft_result = (
        df.with_column("row_num", row_number().over(window_spec))
        .with_column("rank_val", rank().over(window_spec))
        .with_column("dense_rank_val", dense_rank().over(window_spec))
        .sort(["value"], desc=True)
        .collect()
    )

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["value"])


@pytest.mark.parametrize(
    "desc,nulls_first",
    [
        (False, False),
        (False, True),
        (True, False),
        (True, True),
    ],
)
def test_order_by_nulls(desc, nulls_first):
    """Test window functions with ORDER BY NULLS FIRST and NULLS LAST."""
    ids = list(range(1, 7))
    values = [10, None, 20, None, 30, 30]
    data = {
        "id": ids,
        "value": values,
    }
    df = daft.from_pydict(data)
    catalog = SQLCatalog({"test_data": df})
    desc_str = "DESC" if desc else "ASC"
    nulls_first_str = "NULLS FIRST" if nulls_first else "NULLS LAST"
    sql_result = daft.sql(
        f"""
        SELECT
            id,
            value,
            ROW_NUMBER() OVER(ORDER BY value {desc_str} {nulls_first_str}) AS row_num,
            RANK() OVER(ORDER BY value {desc_str} {nulls_first_str}) AS rank_val,
            DENSE_RANK() OVER(ORDER BY value {desc_str} {nulls_first_str}) AS dense_rank_val
        FROM test_data
        ORDER BY id
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().order_by("value", desc=desc, nulls_first=nulls_first)

    daft_result = (
        df.with_column("row_num", row_number().over(window_spec))
        .with_column("rank_val", rank().over(window_spec))
        .with_column("dense_rank_val", dense_rank().over(window_spec))
        .sort(["id"], desc=True)
        .collect()
    )

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["id"])


def test_order_by_only_multiple_columns():
    """Test window functions with ORDER BY multiple columns."""
    df = daft.from_pydict(
        {
            "x": [1, 1, 1, 2, 2, 2],
            "y": [1, 2, 3, 1, 2, 3],
        }
    )

    catalog = SQLCatalog({"test_data": df})

    sql_result = daft.sql(
        """
        SELECT
            x,
            y,
            ROW_NUMBER() OVER(ORDER BY x, y) AS row_num,
            RANK() OVER(ORDER BY x, y) AS rank_val,
            DENSE_RANK() OVER(ORDER BY x, y) AS dense_rank_val
        FROM test_data
        ORDER BY x, y
    """,
        catalog=catalog,
    ).collect()

    window_spec = Window().order_by(["x", "y"])

    daft_result = (
        df.with_column("row_num", row_number().over(window_spec))
        .with_column("rank_val", rank().over(window_spec))
        .with_column("dense_rank_val", dense_rank().over(window_spec))
        .sort(["x", "y"])
        .collect()
    )

    assert_df_equals(sql_result.to_pandas(), daft_result.to_pandas(), sort_key=["x", "y"])
